import logging
import os
import subprocess
import sys
from typing import Optional

import ansys.meshing.prime.internals.config as config
import ansys.meshing.prime.internals.defaults as defaults
import ansys.meshing.prime.internals.utils as utils
from ansys.meshing.prime.internals.client import Client

try:
    import ansys.platform.instancemanagement as pypim
    from simple_upload_server.client import Client as FileClient

    config.set_has_pim(pypim.is_configured())
except ModuleNotFoundError:
    pass

__all__ = ['launch_prime', 'launch_server_process']


def get_install_locations():
    supported_versions = ['232']
    awp_roots = {ver: os.environ.get(f'AWP_ROOT{ver}', '') for ver in supported_versions}
    installed_versions = {
        ver: os.path.join(path, 'meshing', 'Prime')
        for ver, path in awp_roots.items()
        if path and os.path.isdir(path)
    }
    installed_versions = {
        ver: path for ver, path in installed_versions.items() if path and os.path.isdir(path)
    }
    if installed_versions:
        return installed_versions


def get_ansys_prime_server_root():
    available_versions = get_install_locations()
    if not available_versions:
        return None

    version = max(available_versions.keys())
    prime_root = available_versions[version]
    return prime_root


def launch_server_process(
    prime_root: Optional[str] = None,
    ip: str = defaults.ip(),
    port: int = defaults.port(),
    n_procs: Optional[int] = None,
    **kw,
) -> subprocess.Popen:
    '''Launch a server process for Ansys Prime Server.

    Parameters
    ----------
    prime_root: Optional[str]
        Root directory which contains the Ansys Prime Server.
    ip: str
        IP address to start the server at.  The default IP address is 127.0.0.1.
    port: int
        Port at which the server is started. The default port is 50055.
    n_procs: Optional[int]
        When running in distributed mode, specify the number of distributed
        processes to spawn. Process marked as Node 0 will host the GRPC
        server. If None is specified, server will be launched as the only
        process (normal mode).

    Returns
    -------
    subprocess.Popen
        The instance of the subprocess that is launched.

    Raises
    ------
    FileNotFoundError
        When there is an error in the file paths used to launch the server.
    '''
    if prime_root is None:
        prime_root = get_ansys_prime_server_root()
        if prime_root is None:
            raise FileNotFoundError('No valid Ansys Prime Server found to launch.')
    else:  # verify if the file exists
        if not os.path.isdir(prime_root):
            raise FileNotFoundError('Invalid exec_file path.')

    script_ext = 'bat' if os.name == 'nt' else 'sh'
    run_prime_script = f'runPrime.{script_ext}'

    exec_path = os.path.join(prime_root, run_prime_script)
    if not os.path.isfile(exec_path):
        raise FileNotFoundError(f'{run_prime_script} not found in {prime_root}')

    server_args = [exec_path, 'server']

    if kw is None:
        kw = {}

    enable_python_server = kw.get('server', 'release')
    if not isinstance(enable_python_server, str):
        raise ValueError(
            'Hidden option to run python server needs to be a string.'
            'Potential options are: release (default), debug and python'
        )

    enable_python_server = enable_python_server.lower()
    if enable_python_server not in ('python', 'release', 'debug'):
        raise ValueError('server option needs to be one of release, debug or python.')

    if enable_python_server == 'python':
        script_dir = os.path.join(prime_root, 'scripts')
        server_path = os.path.join(script_dir, 'PrimeGRPC.py')
        if not os.path.isfile(server_path):
            raise FileNotFoundError(f'PrimeGRPC.py not found in {script_dir}')
        server_args.append('-python')

    if enable_python_server == 'debug':
        server_args.append('-debug')

    server_args.append(f'--ip={ip}')
    server_args.append(f'--port={port}')
    if n_procs is not None and isinstance(n_procs, int):
        server_args.append(f'-np')
        server_args.append(f'{n_procs}')

    kwargs = {
        'stdin': subprocess.DEVNULL,
    }
    if "PRIME_ENABLE_VERBOSITY" not in os.environ:
        kwargs['stdout'] = subprocess.DEVNULL
        kwargs['stderr'] = subprocess.DEVNULL
    if sys.platform.startswith('win32'):
        kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP

    logging.getLogger('PyPrimeMesh').info('Launching Ansys Prime Server')
    server = subprocess.Popen(server_args, **kwargs)
    return server


def launch_remote_prime(
    version: Optional[str] = None, timeout: float = defaults.connection_timeout()
):
    '''Create a remote instance of Prime server using PyPIM

    This method is used in case of Ansys Labs to create a remote instance of prime.
    This method also creates a file transfer service that is available on Ansys Labs
    '''
    if version is None:
        version = 'latest'

    pim = pypim.connect()
    instance = pim.create_instance(product_name='prime', product_version=version)
    instance.wait_for_ready()

    channel = instance.build_grpc_channel(
        options=[
            ('grpc.max_send_message_length', defaults.max_message_length()),
            ('grpc.max_receive_message_length', defaults.max_message_length()),
        ]
    )

    client = Client(channel=channel, timeout=timeout)
    file_service = FileClient(
        token='token',
        url=instance.services['http-simple-upload-server'].uri,
        headers=instance.services['http-simple-upload-server'].headers,
    )

    client.pim_client = pim
    client.remote_instance = instance
    model = client.model
    model._unfreeze()
    setattr(model, 'file_service', file_service)
    model._freeze()

    return client


def launch_prime(
    prime_root: Optional[str] = None,
    ip: str = defaults.ip(),
    port: int = defaults.port(),
    timeout: float = defaults.connection_timeout(),
    n_procs: Optional[int] = None,
    version: Optional[str] = None,
    **kwargs,
):
    '''Launch an instance of Ansys Prime Server and get a client for it.

    Parameters
    ----------
    prime_root: Optional[str]
        Root directory which contains the Ansys Prime Server.
    ip: str
        IP address to start the server at.  The default IP address is 127.0.0.1.
    port: int
        Port at which the server is started.  The default port is 50055.
    timeout: float
        Maximum time in seconds to wait for the client to connect to the server.
        The default is 10.0 seconds
    n_procs: Optional[int]
        When running in distributed mode, specify the number of distributed
        processes to spawn. Process marked as Node 0 will host the GRPC
        server. If None is specified, server will be launched as the only
        process (normal mode).

    Returns
    -------
    Client
        The instance of the client that is connected to the launched server.

    Raises
    ------
    FileNotFoundError
        When there is an error in file paths used to launch the server.
    ConnectionError
        When there is an error in connecting to the GRPC server.
    '''
    if config.has_pim():
        return launch_remote_prime(version=version, timeout=timeout)

    # Check for port availability on local host
    if ip == defaults.ip():
        port = utils.get_available_local_port(port)

    if 'PYPRIMEMESH_LAUNCH_CONTAINER' in os.environ:
        container_name = 'ansys-prime-server'
        utils.launch_prime_github_container(port=port, name=container_name, version=version)
        config.set_using_container(True)
        client = Client(port=port, timeout=timeout)
        client.container_name = container_name
        return client

    server = launch_server_process(
        prime_root=prime_root, ip=ip, port=port, n_procs=n_procs, **kwargs
    )

    return Client(server_process=server, ip=ip, port=port, timeout=timeout)



""" Auto-generated file. DO NOT MODIFY """
import enum
from typing import Dict, Any, Union, List, Iterable
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.internals import utils
from ansys.meshing.prime.autogen.coreobject import *
import numpy as np

from ansys.meshing.prime.params.primestructs import *

class TransformResults(CoreObject):
    """Results associated with the transformation.
    """
    _default_params = {}

    def __initialize(
            self,
            error_code: ErrorCode):
        self._error_code = ErrorCode(error_code)

    def __init__(
            self,
            model: CommunicationManager=None,
            error_code: ErrorCode = None,
            json_data : dict = None,
             **kwargs):
        """Initializes the TransformResults.

        Parameters
        ----------
        model: Model
            Model to create a TransformResults object with default parameters.
        error_code: ErrorCode, optional
            Error code associated with the failure of operation.
        json_data: dict, optional
            JSON dictionary to create a TransformResults object with provided parameters.

        Examples
        --------
        >>> transform_results = prime.TransformResults(model = model)
        """
        if json_data:
            self.__initialize(
                ErrorCode(json_data["errorCode"] if "errorCode" in json_data else None))
        else:
            all_field_specified = all(arg is not None for arg in [error_code])
            if all_field_specified:
                self.__initialize(
                    error_code)
            else:
                if model is None:
                    raise ValueError("Invalid assignment. Either pass model or specify all properties")
                else:
                    param_json = model._communicator.initialize_params(model, "TransformResults")
                    json_data = param_json["TransformResults"] if "TransformResults" in param_json else {}
                    self.__initialize(
                        error_code if error_code is not None else ( TransformResults._default_params["error_code"] if "error_code" in TransformResults._default_params else ErrorCode(json_data["errorCode"] if "errorCode" in json_data else None)))
        self._custom_params = kwargs
        if model is not None:
            [ model._logger.warning(f'Unsupported argument : {key}') for key in kwargs ]
        [setattr(type(self), key, property(lambda self, key = key:  self._custom_params[key] if key in self._custom_params else None,
        lambda self, value, key = key : self._custom_params.update({ key: value }))) for key in kwargs]
        self._freeze()

    @staticmethod
    def set_default(
            error_code: ErrorCode = None):
        """Set the default values of TransformResults.

        Parameters
        ----------
        error_code: ErrorCode, optional
            Error code associated with the failure of operation.
        """
        args = locals()
        [TransformResults._default_params.update({ key: value }) for key, value in args.items() if value is not None]

    @staticmethod
    def print_default():
        """Print the default values of TransformResults.

        Examples
        --------
        >>> TransformResults.print_default()
        """
        message = ""
        message += ''.join(str(key) + ' : ' + str(value) + '\n' for key, value in TransformResults._default_params.items())
        print(message)

    def _jsonify(self) -> Dict[str, Any]:
        json_data = {}
        if self._error_code is not None:
            json_data["errorCode"] = self._error_code
        [ json_data.update({ utils.to_camel_case(key) : value }) for key, value in self._custom_params.items()]
        return json_data

    def __str__(self) -> str:
        message = "error_code :  %s" % (self._error_code)
        message += ''.join('\n' + str(key) + ' : ' + str(value) for key, value in self._custom_params.items())
        return message

    @property
    def error_code(self) -> ErrorCode:
        """Error code associated with the failure of operation.
        """
        return self._error_code

    @error_code.setter
    def error_code(self, value: ErrorCode):
        self._error_code = value

class TransformParams(CoreObject):
    """Parameters to transform given entities.
    """
    _default_params = {}

    def __initialize(
            self,
            transformation_matrix: Iterable[float]):
        self._transformation_matrix = transformation_matrix if isinstance(transformation_matrix, np.ndarray) else np.array(transformation_matrix, dtype=np.double) if transformation_matrix is not None else None

    def __init__(
            self,
            model: CommunicationManager=None,
            transformation_matrix: Iterable[float] = None,
            json_data : dict = None,
             **kwargs):
        """Initializes the TransformParams.

        Parameters
        ----------
        model: Model
            Model to create a TransformParams object with default parameters.
        transformation_matrix: Iterable[float], optional
            Transformation matrix(4x4) to be used to transform.
        json_data: dict, optional
            JSON dictionary to create a TransformParams object with provided parameters.

        Examples
        --------
        >>> transform_params = prime.TransformParams(model = model)
        """
        if json_data:
            self.__initialize(
                json_data["transformationMatrix"] if "transformationMatrix" in json_data else None)
        else:
            all_field_specified = all(arg is not None for arg in [transformation_matrix])
            if all_field_specified:
                self.__initialize(
                    transformation_matrix)
            else:
                if model is None:
                    raise ValueError("Invalid assignment. Either pass model or specify all properties")
                else:
                    param_json = model._communicator.initialize_params(model, "TransformParams")
                    json_data = param_json["TransformParams"] if "TransformParams" in param_json else {}
                    self.__initialize(
                        transformation_matrix if transformation_matrix is not None else ( TransformParams._default_params["transformation_matrix"] if "transformation_matrix" in TransformParams._default_params else (json_data["transformationMatrix"] if "transformationMatrix" in json_data else None)))
        self._custom_params = kwargs
        if model is not None:
            [ model._logger.warning(f'Unsupported argument : {key}') for key in kwargs ]
        [setattr(type(self), key, property(lambda self, key = key:  self._custom_params[key] if key in self._custom_params else None,
        lambda self, value, key = key : self._custom_params.update({ key: value }))) for key in kwargs]
        self._freeze()

    @staticmethod
    def set_default(
            transformation_matrix: Iterable[float] = None):
        """Set the default values of TransformParams.

        Parameters
        ----------
        transformation_matrix: Iterable[float], optional
            Transformation matrix(4x4) to be used to transform.
        """
        args = locals()
        [TransformParams._default_params.update({ key: value }) for key, value in args.items() if value is not None]

    @staticmethod
    def print_default():
        """Print the default values of TransformParams.

        Examples
        --------
        >>> TransformParams.print_default()
        """
        message = ""
        message += ''.join(str(key) + ' : ' + str(value) + '\n' for key, value in TransformParams._default_params.items())
        print(message)

    def _jsonify(self) -> Dict[str, Any]:
        json_data = {}
        if self._transformation_matrix is not None:
            json_data["transformationMatrix"] = self._transformation_matrix
        [ json_data.update({ utils.to_camel_case(key) : value }) for key, value in self._custom_params.items()]
        return json_data

    def __str__(self) -> str:
        message = "transformation_matrix :  %s" % (self._transformation_matrix)
        message += ''.join('\n' + str(key) + ' : ' + str(value) for key, value in self._custom_params.items())
        return message

    @property
    def transformation_matrix(self) -> Iterable[float]:
        """Transformation matrix(4x4) to be used to transform.
        """
        return self._transformation_matrix

    @transformation_matrix.setter
    def transformation_matrix(self, value: Iterable[float]):
        self._transformation_matrix = value



'''Configuration utility for the PyPrimeMesh library
'''

from contextlib import contextmanager

__all__ = [
    'enable_optimizing_numpy_arrays',
    'disable_optimizing_numpy_arrays',
    'is_optimizing_numpy_arrays',
    'numpy_array_optimization_enabled',
    'numpy_array_optimization_disabled',
    'set_using_container',
    'using_container',
    'set_has_pim',
    'has_pim',
]

__DEFAULT_USE_BINARY = False
__USING_CONTAINER = False
__HAS_PIM = False


def _optimize_vectors():
    '''Gets the value of flag to optimize vectors.'''
    return __DEFAULT_USE_BINARY


def _set_optimize_vectors(optimize_vectors: bool):
    global __DEFAULT_USE_BINARY
    if not isinstance(optimize_vectors, bool):
        raise ValueError(f'Value to set optimize vectors flag should be Boolean.')

    __DEFAULT_USE_BINARY = optimize_vectors


def enable_optimizing_numpy_arrays():
    '''Enables optimizing numpy arrays over the wire.

    This will allow the library to optimize serialization of numpy arrays
    to allow faster deserialization. This can cause the data transferred
    over the network to increase slightly unless the array is of a very
    large size. Consider enabling this option to improve performance if
    your workflow is not limited by network bandwidth for performance. By
    default, optimization is always turned on.
    '''
    _set_optimize_vectors(True)


def disable_optimizing_numpy_arrays():
    '''Disable optimizing numpy arrays over the wire.

    This will disable serialization optimization for numpy arrays when
    sending them over the wire. Typically, this is needed in case the
    performance is limited by a saturated network. Disabling this flag
    can provide some benefit for data transferred over the wire for
    most arrays. By default, optimization is always turned on.
    '''
    _set_optimize_vectors(False)


def is_optimizing_numpy_arrays():
    '''Query if serialization of numpy arrays is turned on.

    By default, this is always turned on.

    Returns
    -------
    bool
        Boolean flag indicating if optimizing of numpy arrays is enabled.
    '''
    return _optimize_vectors()


@contextmanager
def numpy_array_optimization_enabled():
    '''Context helper to execute code with numpy optimization enabled.'''
    old_value = _optimize_vectors()
    _set_optimize_vectors(True)
    yield
    _set_optimize_vectors(old_value)


@contextmanager
def numpy_array_optimization_disabled():
    '''Context helper to execute code with numpy optimization disabled.'''
    old_value = _optimize_vectors()
    _set_optimize_vectors(False)
    yield
    _set_optimize_vectors(old_value)


def set_using_container(value: bool):
    global __USING_CONTAINER
    __USING_CONTAINER = value
    return __USING_CONTAINER


def using_container():
    return __USING_CONTAINER


def set_has_pim(value: bool):
    global __HAS_PIM
    __HAS_PIM = value
    return __HAS_PIM


def has_pim():
    return __HAS_PIM



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class Wrapper(CoreObject):
    """Provide operations to generate surface mesh using wrapper technology.

    """

    def __init__(self, model: CommunicationManager):
        """ Initialize Wrapper """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::Wrapper/Construct"
        args = {"ModelID" : model._object_id , "MaxID" : -1 }
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for Wrapper. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for Wrapper. """
        command_name = "PrimeMesh::Wrapper/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def wrap(self, wrapper_control_id : int, params : WrapParams) -> WrapResult:
        """ Performs wrapping with specified controls in wrapper control and with provided parameters.


        Parameters
        ----------
        wrapper_control_id : int
            Id of wrapper control.
        params : WrapParams
            Wrap Parameters.

        Returns
        -------
        WrapResult
            Returns the WrapResult.


        Examples
        --------
        >>> results = wrapper.wrap(wrapper_control_id, params)

        """
        args = {"wrapper_control_id" : wrapper_control_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::Wrapper/Wrap"
        self._model._print_logs_before_command("wrap", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("wrap", WrapResult(model = self._model, json_data = result))
        return WrapResult(model = self._model, json_data = result)

    def improve_quality(self, part_id : int, params : WrapperImproveQualityParams) -> WrapperImproveResult:
        """ Improve the surface quality and resolve connectivity issues like intersections, multi, free, spikes, point contacts and so on.


        Parameters
        ----------
        part_id : int
            Id of the part.
        params : WrapperImproveQualityParams
            Wrapper improve quality parameters.

        Returns
        -------
        WrapperImproveResult
            Return the Wrapper improve result.


        Examples
        --------
        >>> result = wrapper.improve_quality(part_id, params)

        """
        args = {"part_id" : part_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::Wrapper/ImproveQuality"
        self._model._print_logs_before_command("improve_quality", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("improve_quality", WrapperImproveResult(model = self._model, json_data = result))
        return WrapperImproveResult(model = self._model, json_data = result)

    def close_gaps(self, scope : ScopeDefinition, params : WrapperCloseGapsParams) -> WrapperCloseGapsResult:
        """ Close gaps create patching surfaces within the face zonelets specified by scope using gap size.


        Parameters
        ----------
        scope : ScopeDefinition
            Scope definition of face zonelets.
        params : WrapperCloseGapsParams
            Wrapper close gaps parameters.

        Returns
        -------
        WrapperCloseGapsResult
            Returns the WrapperCloseGapsResult.


        Examples
        --------
        >>> result = wrapper.close_gaps(scope, params)

        """
        args = {"scope" : scope._jsonify(),
        "params" : params._jsonify()}
        command_name = "PrimeMesh::Wrapper/CloseGaps"
        self._model._print_logs_before_command("close_gaps", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("close_gaps", WrapperCloseGapsResult(model = self._model, json_data = result))
        return WrapperCloseGapsResult(model = self._model, json_data = result)



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class TopoUtilities(CoreObject):
    """Performs various general surface utilities algorithms. For example, copy zonelets, resolve surface intersections.

    """

    def __init__(self, model: CommunicationManager, part_id: int):
        """ Initialize TopoUtilities  """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::TopoUtilities/Construct"
        args = {"ModelID" : model._object_id , "PartID" : part_id, "MaxID" : -1}
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for TopoUtilities. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for TopoUtilities. """
        command_name = "PrimeMesh::TopoUtilities/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def fill_hole(self, topo_edges : Iterable[int], params : TopoFillHoleParams) -> TopoFillHoleResult:
        """ Fill holes bounded by given topoedges.


        Parameters
        ----------
        topo_edges : Iterable[int]
            Ids of topoedges to be used to find holes.
        params : TopoFillHoleParams
            Parameters to fill holes.

        Returns
        -------
        TopoFillHoleResult
            Return the TopoFillHoleResult.


        Examples
        --------
        >>> results = topo_utils.fill_hole(topo_edges, params)

        """
        args = {"topo_edges" : topo_edges,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::TopoUtilities/FillHole"
        self._model._print_logs_before_command("fill_hole", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("fill_hole", TopoFillHoleResult(model = self._model, json_data = result))
        return TopoFillHoleResult(model = self._model, json_data = result)



from ansys.meshing.prime.autogen.sizecontrol import SizeControl as _SizeControl

# isort: split
from ansys.meshing.prime.autogen.commonstructs import SetNameResults
from ansys.meshing.prime.autogen.sizecontrolstructs import SizeControlSummaryParams


class SizeControl(_SizeControl):
    """Size control is used to compute size field for volumetric surface meshing."""

    def __init__(self, model, id, object_id, name, local=False):
        """__init__(Model self, int id, int object_id, char* name)"""
        _SizeControl.__init__(self, model, id, object_id, name)
        self._model = model
        self._name = name

    def __str__(self) -> str:
        params = SizeControlSummaryParams(model=self._model)
        result = _SizeControl.get_summary(self, params)
        return result.message

    def set_suggested_name(self, name: str) -> SetNameResults:
        """Sets the unique name for the size control based on the given suggested name.


        Parameters
        ----------
        name : str
            Suggested name for the size control.

        Returns
        -------
        SetNameResults
            Returns a name of the size control.


        Examples
        --------
        >>> size_control.set_suggested_name("control1")

        """
        result = _SizeControl.set_suggested_name(self, name)
        self._name = result.assigned_name
        return result

    @property
    def name(self):
        """Gets the name of size control.

        Returns
        -------
        str
            Returns a name of the size control.

        Examples
        --------
        >>> print(size_control.name)

        """
        return self._name



from typing import Iterable, List

# isort: split
from ansys.meshing.prime.autogen.controldata import ControlData as _ControlData

# isort: split
from ansys.meshing.prime.autogen.commonstructs import DeleteResults
from ansys.meshing.prime.autogen.primeconfig import ErrorCode
from ansys.meshing.prime.autogen.prismcontrol import PrismControl
from ansys.meshing.prime.core.periodiccontrol import PeriodicControl
from ansys.meshing.prime.core.sizecontrol import SizeControl
from ansys.meshing.prime.core.volumecontrol import VolumeControl
from ansys.meshing.prime.core.wrappercontrol import WrapperControl
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import SizingType


class ControlData(_ControlData):
    """ControlData acts as a container for all controls (size controls, prism controls,
    wrapper controls, etc).

    """

    def __init__(self, model: CommunicationManager, id: int, object_id: int, name: str):
        """Initializes ControlData"""
        self._model = model
        self._wrapper_controls = []
        self._size_controls = []
        self._prism_controls = []
        self._volume_controls = []
        self._periodic_controls = []
        _ControlData.__init__(self, model, id, object_id, name)

    def get_wrapper_control_by_name(self, name) -> WrapperControl:
        """Gets the wrapper control by name.


        Parameters
        ----------
        name : str
            Name of the wrapper control.

        Returns
        -------
        WrapperControl
            Returns the wrapper control.

        Examples
        --------
        >>> wrapper_control = model.control_data.get_wrapper_control_by_name("wrappercontol-1")

        """
        for wc in self._wrapper_controls:
            if wc.name == name:
                return wc
        return None

    def create_size_control(self, sizing_type: SizingType) -> SizeControl:
        """Creates size control for the given sizing type.


        Parameters
        ----------
        type : SizingType
            Sizing type used to create a size control.

        Returns
        -------
        SizeControl
            Returns the size control.

        Notes
        -----
        An empty size control is created on calling this API.

        Examples
        --------
        >>> size_control = model.control_data.create_size_control(SizingType.Curvature)

        """
        res = _ControlData.create_size_control(self, sizing_type)
        new_size_control = SizeControl(self._model, res[0], res[1], res[2])
        self._size_controls.append(new_size_control)
        return new_size_control

    def create_prism_control(self) -> PrismControl:
        """Creates a PrismControl.


        Returns
        -------
        PrismControl
            Returns PrismControl.


        Examples
        --------
        >>> prism_control = model.control_data.create_prism_control()

        """
        res = _ControlData.create_prism_control(self)
        new_prism_control = PrismControl(self._model, res[0], res[1], res[2])
        self._prism_controls.append(new_prism_control)
        return new_prism_control

    def create_wrapper_control(self) -> WrapperControl:
        """Creates wrapper control with defaults.


        Returns
        -------
        WrapperControl
            Returns the wrapper control.

        Notes
        -----
        A wrapper control with defaults is created on calling this API.

        Examples
        --------
        >>> wrapper_control = model.control_data.create_wrapper_control()

        """
        res = _ControlData.create_wrapper_control(self)
        new_control = WrapperControl(self._model, res[0], res[1], res[2])
        self._wrapper_controls.append(new_control)
        return new_control

    def get_size_control_by_name(self, name: str) -> SizeControl:
        """Gets the size control by name.


        Parameters
        ----------
        name : str
            Name of the size control.

        Returns
        -------
        SizeControl
            Returns the size control.

        Examples
        --------
        >>> size_control = model.control_data.get_size_control_by_name("SizeControl-1")

        """
        for size_control in self._size_controls:
            if size_control.name == name:
                return size_control
        return None

    def get_prism_control_by_name(self, name: str) -> PrismControl:
        """Gets the prism control by name.


        Parameters
        ----------
        name : str
            Name of the prism control.

        Returns
        -------
        PrismControl
            Returns the prism control.

        Examples
        --------
        >>> prism_control = model.control_data.get_prism_control_by_name("PrismControl-1")

        """
        for prism_control in self._prism_controls:
            if prism_control.name == name:
                return prism_control
        return None

    def delete_controls(self, control_ids: Iterable[int]) -> DeleteResults:
        """Deletes the control for the given id.


        Parameters
        ----------
        control_ids : Iterable[int]
            List of control ids.

        Returns
        -------
        DeleteResults
            Returns the DeleteResults.

        Examples
        --------
        >>> results = model.control_data.delete_controls([size_control.id, volume_control.id])

        """
        res = _ControlData.delete_controls(self, control_ids=control_ids)
        if res.error_code == ErrorCode.NOERROR:
            for id in control_ids:
                for size_control in self._size_controls:
                    if size_control.id == id:
                        self._size_controls.remove(size_control)
                        break
                for wrapper_control in self._wrapper_controls:
                    if wrapper_control.id == id:
                        self._wrapper_controls.remove(wrapper_control)
                        break
                for prism_control in self._prism_controls:
                    if prism_control.id == id:
                        self._prism_controls.remove(prism_control)
                        break
                for volume_control in self._volume_controls:
                    if volume_control.id == id:
                        self._volume_controls.remove(volume_control)
                        break
                for periodic_control in self._periodic_controls:
                    if periodic_control.id == id:
                        self._periodic_controls.remove(periodic_control)
                        break
        return res

    def _update_size_controls(self, sc_data: List):
        self._size_controls = [SizeControl(self._model, sc[0], sc[1], sc[2]) for sc in sc_data]

    def _update_prism_controls(self, sc_data: List):
        self._prism_controls = [PrismControl(self._model, sc[0], sc[1], sc[2]) for sc in sc_data]

    def _update_wrapper_controls(self, sc_data: List):
        self._wrapper_controls = [
            WrapperControl(self._model, sc[0], sc[1], sc[2]) for sc in sc_data
        ]

    def _update_volume_controls(self, sc_data: List):
        self._volume_controls = [VolumeControl(self._model, sc[0], sc[1], sc[2]) for sc in sc_data]

    def _update_periodic_controls(self, sc_data: List):
        self._periodic_controls = [
            PeriodicControl(self._model, sc[0], sc[1], sc[2]) for sc in sc_data
        ]

    def create_volume_control(self) -> VolumeControl:
        """Creates the volume control.

        Returns
        -------
        VolumeControl
            Returns the volume control.

        Examples
        --------
        >>> volume_control = model.control_data.create_volume_control()

        """
        res = _ControlData.create_volume_control(self)
        new_control = VolumeControl(self._model, res[0], res[1], res[2])
        self._volume_controls.append(new_control)
        return new_control

    def get_volume_control_by_name(self, name: str) -> VolumeControl:
        """Gets the volume control by name.


        Parameters
        ----------
        name : str
            Name of the volume control.

        Returns
        -------
        VolumeControl
            Returns the volume control.

        Examples
        --------
        >>> volume_control = model.control_data.get_volume_control_by_name("VolumeControl-1")

        """
        for volume_control in self._volume_controls:
            if volume_control.name == name:
                return volume_control
        return None

    @property
    def size_controls(self) -> List[SizeControl]:
        """Get the size controls.

        Returns
        -------
        List[SizeControl]
            Returns the list of size controls.

        Examples
        --------
            >>> size_controls = model.control_data.size_controls
        """
        return self._size_controls

    @property
    def volume_controls(self) -> List[VolumeControl]:
        """Get the volume controls.

        Returns
        -------
        List[VolumeControl]
            Returns the list of volume controls.

        Examples
        --------
            >>> volume_controls = model.control_data.volume_controls
        """
        return self._volume_controls

    @property
    def prism_controls(self) -> List[PrismControl]:
        """Get the prism controls.

        Returns
        -------
        List[PrismControl]
            Returns the list of prism controls.

        Examples
        --------
        >>> prism_control = model.control_data.prism_controls

        """
        return self._prism_controls

    @property
    def wrapper_controls(self) -> List[WrapperControl]:
        """Get the wrapper controls.

        Returns
        -------
        List[WrapperControl]
            Returns the list of wrapper controls.

        Examples
        --------
        >>> wrapper_control = model.control_data.wrapper_controls

        """
        return self._wrapper_controls

    def create_periodic_control(self) -> PeriodicControl:
        """Creates the periodic control.

        Returns
        -------
        PeriodicControl
            Returns the periodic control.

        Examples
        --------
        >>> periodic_control = model.control_data.create_periodic_control()

        """
        res = _ControlData.create_periodic_control(self)
        new_control = PeriodicControl(self._model, res[0], res[1], res[2])
        self._periodic_controls.append(new_control)
        return new_control

    def get_periodic_control_by_name(self, name: str) -> PeriodicControl:
        """Gets the periodic control by name.


        Parameters
        ----------
        name : str
            Name of the periodic control.

        Returns
        -------
        PeriodicControl
            Returns the periodic control.

        Examples
        --------
        >>> periodic_control = model.control_data.get_periodic_control_by_name("PeriodicControl-1")

        """
        for periodic_control in self._periodic_controls:
            if periodic_control.name == name:
                return periodic_control
        return None

    @property
    def periodic_controls(self) -> List[PeriodicControl]:
        """Get the periodic controls.

        Returns
        -------
        List[PeriodicControl]
            Returns the list of periodic controls.

        Examples
        --------
            >>> periodic_controls = model.control_data.periodic_controls
        """
        return self._periodic_controls



import enum
import os
from typing import Iterable, List

import ansys.meshing.prime as prime

from .scope import SurfaceScope, VolumeScope
from .utils import check_name_pattern


class LabelToZoneMethod(enum.IntEnum):
    """Method to create zones from labels."""

    SIMPLE = 0
    """
    Simple method to create zones from labels.
    Entities are queried using labels and zones created.
    """


class Mesh:
    """Mesh is one of the classes in Lucid API.

    This class is meant for beginners to meshing.
    It also serves as a tutorial for commonly used meshing workflows.

    The following functionalities are provided by this class:
    * Surface meshing with constant and variable sizing with triangle or quad dominant mesh.
    * Volume meshing with prism, tetrahedral and polyhedral elements.
    * Surface wrapping.
    * Helper method to create zones from labels.
    * Helper methods for reading and writing files.
    """

    def __init__(self, model: prime.Model):
        """Initialize mesher using model.

        Parameters
        ----------
        model : prime.Model
            Model on which the methods work.
        """
        self._model = model
        self._logger = model.python_logger

    def read(
        self, file_name: str, append: bool = False, cad_reader_route: prime.CadReaderRoute = None
    ):
        """Read or import files of different formats based on file extension.

        PyPrimeMesh's native file format has extension pmdat.
        The method supports the following:
        * Reading PyPrimeMesh's native file format.
        * Importing various CAD Formats.
        * Importing Fluent Meshing's msh file.
        * Importing Fluent cas file.
        * Importing MAPDL cdb files.

        Parameters
        ----------
        file_name : str
            Path to file to be read or imported.

        append : bool
            Set it to True to append instead of overwrite.

        cad_reader_route : prime.CadReaderRoute
            Route of CadReader.

        """
        filename, fileext = os.path.splitext(file_name)
        if fileext == ".msh" or file_name[-7:] == ".msh.gz":
            prime.FileIO(self._model).import_fluent_meshing_meshes(
                [file_name], prime.ImportFluentMeshingMeshParams(self._model, append=append)
            )
        elif fileext == ".cas":
            prime.FileIO(self._model).import_fluent_case(
                file_name, prime.ImportFluentCaseParams(self._model, append=append)
            )
        elif fileext == ".cdb":
            prime.FileIO(self._model).import_mapdl_cdb(
                file_name, prime.ImportMapdlCdbParams(self._model, append=append)
            )
        elif fileext == ".pmdat":
            prime.FileIO(self._model).read_pmdat(
                file_name, prime.FileReadParams(self._model, append=append)
            )
        else:
            prime.FileIO(self._model).import_cad(
                file_name,
                prime.ImportCadParams(
                    self._model, append=append, cad_reader_route=cad_reader_route
                ),
            )

    def write(self, file_name: str):
        """Write or export files of different formats based on file extension.

        PyPrimeMesh's native file format has extension pmdat.
        The method supports the following:
        * Writing PyPrimeMesh's native file format.
        * Exporting Fluent Meshing's msh file.
        * Exporting Fluent cas file.
        * Exporting MAPDL cdb files.

        Parameters
        ----------
        file_name : str
            Path of file to be written or exported.

        """
        filename, fileext = os.path.splitext(file_name)
        if fileext == ".cdb":
            prime.FileIO(self._model).export_mapdl_cdb(
                file_name, prime.ExportMapdlCdbParams(self._model)
            )
        elif fileext == ".cas":
            prime.FileIO(self._model).export_fluent_case(
                file_name, prime.ExportFluentCaseParams(self._model)
            )
        elif fileext == ".msh" or file_name[-7:] == ".msh.gz":
            prime.FileIO(self._model).export_fluent_meshing_mesh(
                file_name, prime.ExportFluentMeshingMeshParams(self._model)
            )
        else:
            prime.FileIO(self._model).write_pmdat(file_name, prime.FileWriteParams(self._model))

    def create_zones_from_labels(
        self,
        label_expression: str = None,
        conversion_method: LabelToZoneMethod = LabelToZoneMethod.SIMPLE,
    ):
        """Create zones from labels.

        When exporting to various solvers, zones play a very important role.
        Zones are where material properties and boundary conditions
        can be set in respective solvers.

        Zones allow downstream setting of
        boundary conditions or material properties.

        Zone names in PyPrimeMesh are translated into equivalent concepts in solver.
        Currently, only one method is available to convert zone to label.
        Currently, only face zones are created.

        The method finds the entities by labels and then adds them to the
        zone with the same name as label.

        If no label_expression is provided then all labels will be flattened to
        create zones.  Label names will be combined if overlaps occur and
        separate zones created.

        For example, if "LabelA" and "LabelB" had overlapping TopoFaces then
        the following three zones would be created;
        "LabelA", "LabelB" and "LabelA_LabelB" for the overlap.

        Parameters
        ----------
        label_expression : str
            Expression for labels to be converted to zones.

        conversion_method : LabelToZoneMethod
            Method used to convert label to zones.

        Examples
        --------
        >>> from ansys.meshing.prime import lucid
        >>> mesh_util = lucid.Mesh(model)
        >>> mesh_util.create_zones_from_labels()

        """
        if conversion_method != LabelToZoneMethod.SIMPLE:
            self._logger.error("Invalid label to zone conversion method")

        if label_expression:
            labels = []
            for part in self._model.parts:
                lbls = part.get_labels()
                for l in lbls:
                    if l not in labels and check_name_pattern(label_expression, l):
                        labels.append(l)

            for label in labels:
                face_zone = None
                for part in self._model.parts:
                    topo_faces = part.get_topo_faces()
                    if len(topo_faces) > 0:
                        in_topo_faces = part.get_topo_faces_of_label_name_pattern(
                            label, prime.NamePatternParams(model=self._model)
                        )
                        if len(in_topo_faces) > 0:
                            if face_zone == None:
                                face_zone = self._model.create_zone(label, prime.ZoneType.FACE)
                            part.add_topo_entities_to_zone(face_zone.zone_id, in_topo_faces)
                    else:
                        in_face_zonelets = part.get_face_zonelets_of_label_name_pattern(
                            label, prime.NamePatternParams(model=self._model)
                        )
                        if len(in_face_zonelets) > 0:
                            if face_zone == None:
                                face_zone = self._model.create_zone(label, prime.ZoneType.FACE)
                            part.add_zonelets_to_zone(face_zone.zone_id, in_face_zonelets)
        else:
            labels = []
            for part in self._model.parts:
                labels += part.get_labels()
            faces_of_label = {}
            labels_of_face = {}
            label_zone_definitions = {}
            face_zones = []
            all_topo = []
            for part in self._model.parts:
                for face in part.get_topo_faces():
                    labels_of_face[face] = []
                for face in part.get_face_zonelets():
                    labels_of_face[face] = []
            for label in labels:
                faces = []
                faces_of_label[label] = []
                for part in self._model.parts:
                    for zone in part.get_face_zones():
                        face_zones.append(self._model.get_zone_name(zone))
                    topo_faces = part.get_topo_faces()
                    all_topo += topo_faces
                    if topo_faces:
                        faces += part.get_topo_faces_of_label_name_pattern(
                            label_name_pattern=label,
                            name_pattern_params=prime.NamePatternParams(self._model),
                        )
                    else:
                        faces += part.get_face_zonelets_of_label_name_pattern(
                            label_name_pattern=label,
                            name_pattern_params=prime.NamePatternParams(self._model),
                        )
                if faces:
                    faces_of_label[label] += faces
                    [labels_of_face[face].append(label) for face in faces]
            for face in labels_of_face:
                name_exists = False
                for zone_name in label_zone_definitions:
                    if set(zone_name) == set(labels_of_face[face]):
                        name_exists = True
                        label_zone_definitions[zone_name].append(face)
                if not name_exists:
                    label_zone_definitions['_'.join(labels_of_face[face])] = [face]
            # remove empty labels
            if "" in label_zone_definitions:
                label_zone_definitions.pop("")
            self._logger.info("Labels to zones: " + str(label_zone_definitions))
            for zone_name in label_zone_definitions:
                if zone_name not in face_zones:
                    self._model.create_zone(zone_name, prime.ZoneType.FACE)
                for face in label_zone_definitions[zone_name]:
                    for part in self._model.parts:
                        if face in all_topo and face in part.get_topo_faces():
                            part.add_topo_entities_to_zone(
                                topo_entities=[face],
                                zone_id=self._model.get_zone_by_name(zone_name=zone_name),
                            )
                        elif face not in all_topo and face in part.get_face_zonelets():
                            part.add_zonelets_to_zone(
                                zonelets=[face],
                                zone_id=self._model.get_zone_by_name(zone_name=zone_name),
                            )

    def merge_parts(self, parts_expression: str = "*", new_name: str = "merged_part"):
        """Merges given parts into one.

        Parameters
        ----------
        parts_expression : str
            Expression of parts to be merged.
        new_name : str
            New part name for the merged part.

        """
        part_ids = []
        for part in self._model.parts:
            if check_name_pattern(parts_expression, part.get_name()):
                part_ids.append(part.id)

        self._model.merge_parts(
            part_ids=part_ids,
            params=prime.MergePartsParams(self._model, merged_part_suggested_name=new_name),
        )

    def delete_topology(self, parts_expression: str = "*", delete_edges: bool = True):
        """Delete topoentities of part.

        Parameters
        ----------
        parts_expression : str
            Expression of parts whose topology needs to be deleted.
        delete_edges : bool
            Check whether to delete the edges or not.

        """
        for part in self._model.parts:
            if check_name_pattern(parts_expression, part.get_name()):
                part.delete_topo_entities(
                    prime.DeleteTopoEntitiesParams(
                        self._model, delete_geom_zonelets=True, delete_mesh_zonelets=False
                    )
                )
                if delete_edges:
                    part.delete_zonelets(part.get_edge_zonelets())

    def __surface_mesh_on_active_sf(self, generate_quads: bool, scope: SurfaceScope):
        part_ids = scope.get_parts(self._model)
        for part_id in part_ids:
            topofaces = scope.get_topo_faces(self._model, part_id)
            surfer = prime.Surfer(self._model)
            params = prime.SurferParams(
                self._model,
                size_field_type=prime.SizeFieldType.VOLUMETRIC,
                generate_quads=generate_quads,
            )
            if len(topofaces) > 0:
                surfer.mesh_topo_faces(part_id=part_id, topo_faces=topofaces, params=params)
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                if len(tf) > 0:
                    te = []
                    surfer.remesh_face_zonelets(part_id, tf, te, params)

    def __variable_size_surface_mesh(
        self, min_size: float, max_size: float, generate_quads: bool, scope: SurfaceScope
    ):
        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(model=self._model, min=min_size, max=max_size, growth_rate=1.2)
        )
        surfer = prime.Surfer(self._model)
        params = prime.SurferParams(
            self._model,
            size_field_type=prime.SizeFieldType.VOLUMETRIC,
            generate_quads=generate_quads,
        )
        part_ids = scope.get_parts(self._model)
        for part_id in part_ids:
            part = self._model.get_part(part_id)
            topofaces = scope.get_topo_faces(self._model, part_id)
            if len(topofaces) > 0:
                sizecontrol1 = self._model.control_data.create_size_control(
                    prime.SizingType.CURVATURE
                )
                sizecontrol1.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model, min=min_size, max=max_size, growth_rate=1.2
                    )
                )
                sizecontrol1.set_scope(
                    prime.ScopeDefinition(
                        model=self._model,
                        entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
                        part_expression=part.name,
                    )
                )
                sizefield = prime.SizeField(model=self._model)
                res = sizefield.compute_volumetric(
                    size_control_ids=[sizecontrol1.id],
                    volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                        model=self._model, enable_multi_threading=False
                    ),
                )
                surfer.mesh_topo_faces(part_id=part.id, topo_faces=topofaces, params=params)
                self._model.delete_volumetric_size_fields([res.size_field_id])
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                if len(tf) > 0:
                    te = []
                    sizecontrol1 = self._model.control_data.create_size_control(
                        prime.SizingType.CURVATURE
                    )
                    sizecontrol1.set_curvature_sizing_params(
                        prime.CurvatureSizingParams(
                            model=self._model, min=min_size, max=max_size, growth_rate=1.2
                        )
                    )
                    sizecontrol1.set_scope(
                        prime.ScopeDefinition(
                            model=self._model,
                            entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
                            part_expression=part.name,
                        )
                    )
                    sizefield = prime.SizeField(model=self._model)
                    res = sizefield.compute_volumetric(
                        size_control_ids=[sizecontrol1.id],
                        volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                            model=self._model, enable_multi_threading=False
                        ),
                    )
                    surfer.remesh_face_zonelets(part.id, tf, te, params)
                    self._model.delete_volumetric_size_fields([res.size_field_id])

    def __constant_size_surface_mesh(
        self, min_size: float, generate_quads: bool, scope: SurfaceScope
    ):
        part_ids = scope.get_parts(self._model)
        surfer = prime.Surfer(self._model)
        for part_id in part_ids:
            topofaces = scope.get_topo_faces(self._model, part_id)
            params = prime.SurferParams(
                self._model, constant_size=min_size, generate_quads=generate_quads
            )
            if len(topofaces) > 0:
                surfer.mesh_topo_faces(part_id=part_id, topo_faces=topofaces, params=params)
            else:
                tf = scope.get_face_zonelets(self._model, part_id)
                te = []
                surfer.remesh_face_zonelets(part_id, tf, te, params)

    def surface_mesh(
        self,
        min_size: float = None,
        max_size: float = None,
        generate_quads: bool = False,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate Surface mesh on the given scope.

        This method generates surface mesh on the given scope.
        The method is used to generate constant or variable size surface mesh.
        The method supports generating quad dominant or triangular elements.

        If min size and max size are provided, variable size mesh is generated
        between min size and max size by applying sizes based on curvature.

        If either of min size or max size is provided, constant size mesh
        will be generated with the provided size.

        If neither of min size or max size is provided, global max setting is used to
        generate a constant size mesh.

        Parameters
        ----------
        min_size : float
            Minimum edge length of the mesh.

        max_size : float
            Maximum edge length of the mesh.

        generate_quads : bool
            Generate quad dominant mesh or all triangular mesh.

        scope : SurfaceScope
            Scope for generating surface mesh.

        """
        if min_size == None and max_size == None:
            global_sizing = self._model.get_global_sizing_params()
            self._logger.warning(
                "Min and Max size not provided. Using max global size "
                + str(global_sizing.max)
                + " and min global size "
                + str(global_sizing.min)
            )
            self.__variable_size_surface_mesh(
                min_size=global_sizing.min,
                max_size=global_sizing.max,
                generate_quads=generate_quads,
                scope=scope,
            )
        elif min_size == None or max_size == None:
            if min_size is None:
                self.__constant_size_surface_mesh(
                    min_size=max_size, generate_quads=generate_quads, scope=scope
                )
            else:
                self.__constant_size_surface_mesh(
                    min_size=min_size, generate_quads=generate_quads, scope=scope
                )
        else:
            self.__variable_size_surface_mesh(
                min_size=min_size, max_size=max_size, generate_quads=generate_quads, scope=scope
            )

    def create_constant_size_control(
        self,
        control_name: str = "size_control",
        size: float = 1.0,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate constant size control on the given scope.

        This method generates constant size control on the given scope.

        Parameters
        ----------
        control_name : str
            Name of the control.

        size : float
            Constant edge length of the mesh.

        scope : SurfaceScope
            Scope for creating size control.

        """
        global_sizes = self._model.get_global_sizing_params()
        if global_sizes.max < size:
            global_sizes.max = size
        if global_sizes.min > size or global_sizes.min <= 0:
            global_sizes.min = size
        self._model.set_global_sizing_params(global_sizes)
        sizecontrol = self._model.control_data.create_size_control(prime.SizingType.SOFT)
        sizecontrol.set_soft_sizing_params(
            prime.SoftSizingParams(
                model=self._model, max=size, growth_rate=global_sizes.growth_rate
            )
        )
        sizecontrol.set_scope(scope.get_scope_definition(self._model))
        sizecontrol.set_suggested_name(control_name)

    def create_curvature_size_control(
        self,
        control_name: str = "size_control",
        min: float = 1.0,
        max: float = 2.0,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate curvature size control on the given scope.

        This method generates curvature size control on the given scope.

        Parameters
        ----------
        control_name : str
            Name of the control.

        min : float
            Min edge length of the mesh.

        max : float
            Max edge length of the mesh.

        scope : SurfaceScope
            Scope for creating size control.

        """
        global_sizes = self._model.get_global_sizing_params()
        if max < min:
            t = max
            max = min
            min = t
        if global_sizes.max < max:
            global_sizes.max = max
        if global_sizes.min > min or global_sizes.min <= 0:
            global_sizes.min = min
        self._model.set_global_sizing_params(global_sizes)
        sizecontrol = self._model.control_data.create_size_control(prime.SizingType.CURVATURE)
        sizecontrol.set_curvature_sizing_params(
            prime.CurvatureSizingParams(
                model=self._model, min=min, max=max, growth_rate=global_sizes.growth_rate
            )
        )
        sizecontrol.set_scope(scope.get_scope_definition(self._model))
        sizecontrol.set_suggested_name(control_name)

    def surface_mesh_with_size_controls(
        self,
        size_control_names: str = "*",
        generate_quads: bool = False,
        scope: SurfaceScope = SurfaceScope(),
    ):
        """Generate Surface mesh on the given scope with the given size controls.

        This method generates surface mesh on the given scope.
        The method is used to generate surface mesh using the given size controls.
        The method supports generating quad dominant or triangular elements.

        Parameters
        ----------
        size_control_names : str
            Name pattern for the size controls.

        generate_quads : bool
            Generate quad dominant mesh or all triangular mesh.

        scope : SurfaceScope
            Scope for generating surface mesh.

        """
        sizefield = prime.SizeField(model=self._model)
        s_control_ids = []
        s_controls = self._model.control_data.size_controls
        for control in s_controls:
            if check_name_pattern(size_control_names, control.name):
                s_control_ids.append(control.id)
        if len(s_control_ids) > 0:
            size_field_res = sizefield.compute_volumetric(
                size_control_ids=s_control_ids,
                volumetric_sizefield_params=prime.VolumetricSizeFieldComputeParams(
                    model=self._model, enable_multi_threading=False
                ),
            )
            self.__surface_mesh_on_active_sf(generate_quads=generate_quads, scope=scope)
            self._model.delete_volumetric_size_fields([size_field_res.size_field_id])

    def connect_faces(
        self,
        part_expression: str = "*",
        face_labels: str = "*",
        target_face_labels: str = "*",
        tolerance: float = 0.05,
    ):
        """Connect face zonelets with given label name pattern within the given tolerance.

        This method is used to connect face zonelets of given label name pattern to the
        face zonelets given by target face labels within a given tolerance. Connect happens within
        part. Face zonelets of a part are connected with face zonelets of the same part only.

        Parameters
        ----------
        part_expression: str
            Name pattern of parts used for connecting.
        face_labels: str
            Name pattern of face labels used for connecting.
        target_face_labels: str
            Name pattern of face labels with which you want to connect.
        tolerance: float
            Tolerance used for connection.

        """
        name_pattern_param = prime.NamePatternParams(self._model)
        connect = prime.Connect(self._model)
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                join_faces_source = part.get_face_zonelets_of_label_name_pattern(
                    face_labels, name_pattern_param
                )
                join_faces_target = part.get_face_zonelets_of_label_name_pattern(
                    target_face_labels, name_pattern_param
                )
                if len(join_faces_source) > 0 and len(join_faces_target) > 0:
                    connect.intersect_face_zonelets(
                        part.id,
                        face_zonelet_ids=join_faces_source,
                        with_face_zonelet_ids=join_faces_target,
                        params=prime.IntersectParams(self._model, tolerance=tolerance),
                    )
                    connect.join_face_zonelets(
                        part.id,
                        face_zonelet_ids=join_faces_source,
                        with_face_zonelet_ids=join_faces_target,
                        params=prime.JoinParams(self._model, tolerance=tolerance),
                    )

    def compute_volumes(self, part_expression: str = "*", create_zones_per_volume: bool = True):
        """
        Computes volumes in the parts defined by the part expression

        Parameters
        ----------
        part_expression : str
            Expression of parts where topology needs to be deleted.

        create_zones_per_volume : bool
            Creates volume zones for each volume when True.

        """
        create_zones_type = prime.CreateVolumeZonesType.NONE
        if create_zones_per_volume:
            create_zones_type = prime.CreateVolumeZonesType.PERVOLUME
        params = prime.ComputeVolumesParams(model=self._model, create_zones_type=create_zones_type)
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                if len(part.get_topo_faces()) > 0:
                    part.compute_topo_volumes(params=params)
                else:
                    part.compute_closed_volumes(params=params)

    def delete_topology(self, part_expression: str = "*", delete_edges: bool = True):
        """
        Deletes topology in the given part.

        This method can be used to delete topology in the parts defined by the part expression.
        If delete_edges is set to True, edge zonelets will be deleted.

        Parameters
        ----------
        part_expression : str
            Expression of parts where topology needs to be deleted.

        delete_edges : bool
            Edge zonelets are deleted when true.

        """
        for part in self._model.parts:
            if check_name_pattern(part_expression, part.name):
                part.delete_topo_entities(
                    prime.DeleteTopoEntitiesParams(
                        model=self._model, delete_geom_zonelets=True, delete_mesh_zonelets=False
                    )
                )
                if delete_edges:
                    part.delete_zonelets(part.get_edge_zonelets())

    def __create_cap(self, part_id: int, face_zonelets: Iterable[int]):
        su = prime.SurfaceUtilities(model=self._model)
        res = su.create_cap_on_face_zonelets(
            part_id=part_id,
            face_zonelets=face_zonelets,
            params=prime.CreateCapParams(model=self._model),
        )
        return res.created_face_zonelets

    def __compute_flow_volume(
        self, part_id: int, face_zonelets: Iterable[int], flow_volume_zone_name: str = "flow_volume"
    ):
        part = self._model.get_part(part_id)
        if part:
            params = prime.ExtractVolumesParams(
                model=self._model, create_zone=True, suggested_zone_name=flow_volume_zone_name
            )
            res = part.extract_volumes(face_zonelets=face_zonelets, params=params)

    def create_flow_volume(
        self, flow_volume_zone_name: str = "flow_volume", cap_scope: SurfaceScope = SurfaceScope()
    ):
        """Create flow volume by the given face labels defining the boundary of the volume.

        This method creates flow volumes for the given faces defining the boundary of the volume.

        Parameters
        ----------
        flow_volume_zone_name: str
            Suggested name for the volume zone of the created flow volume.
        cap_scope: SurfaceScope
            Scope defining the face zonelets where cap for flow volume needs to be created.

        """
        parts = cap_scope.get_parts(self._model)
        for part in parts:
            caps = self.__create_cap(part, cap_scope.get_face_zonelets(self._model, part))
            if len(caps) > 0:
                self.__compute_flow_volume(part, caps, flow_volume_zone_name)

    def __create_volume_controls(self, part: prime.Part, scope: VolumeScope) -> Iterable[int]:
        volume_control_ids = []
        if scope._evaluation_type == prime.ScopeEvaluationType.ZONES:
            volume_zones_all = part.get_volume_zones()
            volume_zones_to_mesh = []
            for volume_zone in volume_zones_all:
                if check_name_pattern(
                    scope._entity_expression, self._model.get_zone_name(volume_zone)
                ):
                    volume_zones_to_mesh.append(volume_zone)

            volume_zones_to_avoid = [v for v in volume_zones_all if v not in volume_zones_to_mesh]
            scope_str = ", ".join([self._model.get_zone_name(v) for v in volume_zones_to_avoid])

            if len(scope_str) > 0:
                volume_control_param_dead = prime.VolumeControlParams(
                    model=self._model, cell_zonelet_type=prime.CellZoneletType.DEAD
                )
                volume_control_scope_dead = prime.ScopeDefinition(
                    model=self._model,
                    part_expression=part.name,
                    label_expression="",
                    zone_expression=scope_str,
                    entity_type=prime.ScopeEntity.VOLUME,
                    evaluation_type=prime.ScopeEvaluationType.ZONES,
                )
                volume_control_dead = self._model.control_data.create_volume_control()
                volume_control_dead.set_params(volume_control_params=volume_control_param_dead)
                volume_control_dead.set_scope(scope=volume_control_scope_dead)
                volume_control_ids.append(volume_control_dead.id)

            volume_control_param_fluid = prime.VolumeControlParams(
                model=self._model, cell_zonelet_type=prime.CellZoneletType.FLUID
            )
            volume_control_scope_fluid = prime.ScopeDefinition(
                model=self._model,
                part_expression=part.name,
                label_expression="",
                zone_expression=scope._entity_expression,
                entity_type=prime.ScopeEntity.VOLUME,
                evaluation_type=prime.ScopeEvaluationType.ZONES,
            )
            volume_control_fluid = self._model.control_data.create_volume_control()
            volume_control_fluid.set_params(volume_control_params=volume_control_param_fluid)
            volume_control_fluid.set_scope(scope=volume_control_scope_fluid)
            volume_control_ids.append(volume_control_fluid.id)

        return volume_control_ids

    def volume_mesh(
        self,
        volume_fill_type: prime.VolumeFillType = prime.VolumeFillType.TET,
        quadratic: bool = False,
        prism_layers: int = None,
        prism_surface_expression: str = "*",
        prism_volume_expression: str = "*",
        growth_rate: float = 1.2,
        scope: VolumeScope = VolumeScope(),
    ):
        """Generate Volume mesh on the model.

        This method can be used to generate volume mesh on the entire model.
        If prism layers parameter is set, prism layers are generated.

        Parameters
        ----------
        volume_fill_type : prime.VolumeFillType
            Type of volume elements to be generated.

        quadratic : bool
            Option to generate quadratic mesh. It is not supported with parallel meshing.
            It is only supported with pure tetrahedral mesh.

        prism_layers : int
            Number of prism layers to grow.

        prism_surface_expression : str
            Facezonelets or topofaces from which prisms are grown.
            Default is to grow from all surfaces.

        prism_volume_expression : str
            Volumes or topovolumes into which prisms are grown.
            The expression evaluates to zone names and volumes
            or topovolumes are queried based on zones evaluated.

        growth_rate : float
            Prism growth rate.

        scope : VolumeScope
            Scope of volumes to be meshed.

        """
        automesh_params = prime.AutoMeshParams(model=self._model)
        automesh_params.volume_fill_type = volume_fill_type
        if quadratic:
            automesh_params.tet = prime.TetParams(self._model, True)
        automesh = prime.AutoMesh(self._model)
        for part in self._model.parts:
            if check_name_pattern(scope._part_expression, part.name):
                try:
                    prism_control = None
                    volume_control_ids = self.__create_volume_controls(part, scope)
                    if len(volume_control_ids) > 0:
                        automesh_params.volume_control_ids = volume_control_ids
                    if prism_layers:
                        prism_control = self._model.control_data.create_prism_control()
                        prism_control.set_surface_scope(
                            prime.ScopeDefinition(
                                self._model,
                                part_expression=part.name,
                                label_expression=prism_surface_expression,
                            )
                        )
                        prism_control.set_growth_params(
                            prime.PrismControlGrowthParams(
                                self._model, n_layers=prism_layers, growth_rate=growth_rate
                            )
                        )
                        prism_control.set_volume_scope(
                            prime.ScopeDefinition(
                                self._model,
                                entity_type=prime.ScopeEntity.VOLUME,
                                evaluation_type=prime.ScopeEvaluationType.ZONES,
                                part_expression=part.name,
                                zone_expression=prism_volume_expression,
                            )
                        )
                        automesh_params.prism_control_ids = [prism_control.id]

                    automesh.mesh(part.id, automesh_params=automesh_params)
                    if prism_control:
                        self._model.control_data.delete_controls([prism_control.id])
                    if len(volume_control_ids) > 0:
                        self._model.control_data.delete_controls(volume_control_ids)
                except:
                    self._logger.info(part.name + " not volume meshed.")

    def __setup_sizing_for_wrapper(
        self,
        min_size: float,
        max_size: float,
        growth_rate: float,
        normal_angle: float,
        elements_per_gap: float,
        global_size_controls: List,
        geodesic_global_size_controls: List,
        scope: prime.ScopeDefinition,
        global_sf: prime.GlobalSizingParams,
        wrap_size_controls: List[prime.SizeControl],
    ):
        constant_size = None
        if min_size and not max_size:
            max_size = min_size
            constant_size = min_size
        elif max_size and not min_size:
            min_size = max_size
            constant_size = max_size

        if not max_size and not min_size:
            max_size = global_sf.max
            min_size = global_sf.min
        if not growth_rate:
            growth_rate = global_sf.growth_rate
        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(
                self._model, min=min_size, max=max_size, growth_rate=growth_rate
            )
        )
        self._logger.info(
            "Min size: "
            + str(min_size)
            + ", Max size: "
            + str(max_size)
            + ", Growth rate: "
            + str(growth_rate)
        )

        is_geodesic = self.__is_size_field_type_geodesic(wrap_size_controls)
        if not constant_size:
            if normal_angle:
                curv_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.CURVATURE
                )
                curv_size_control.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model,
                        max=max_size,
                        min=min_size,
                        growth_rate=growth_rate,
                        normal_angle=normal_angle,
                    )
                )
                curv_size_control.set_scope(scope)
                global_size_controls.append(curv_size_control)
                if is_geodesic or elements_per_gap:
                    curv_size_control = self._model.control_data.create_size_control(
                        sizing_type=prime.SizingType.CURVATURE
                    )
                    curv_size_control.set_curvature_sizing_params(
                        prime.CurvatureSizingParams(
                            model=self._model,
                            max=max_size,
                            min=min_size,
                            growth_rate=growth_rate,
                            normal_angle=normal_angle,
                        )
                    )
                    curv_size_control.set_scope(scope)
                    geodesic_global_size_controls.append(curv_size_control)
            else:
                if elements_per_gap == None and not wrap_size_controls:
                    prime.PrimeRuntimeError(
                        "Error: No size functions or settings provided.\
                        No wrap done."
                    )
                    return
            if elements_per_gap:
                prox_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.PROXIMITY
                )
                prox_size_control.set_proximity_sizing_params(
                    prime.ProximitySizingParams(
                        model=self._model,
                        elements_per_gap=elements_per_gap,
                        max=max_size,
                        min=min_size,
                        growth_rate=growth_rate,
                        ignore_orientation=True,
                        ignore_self_proximity=False,
                    )
                )
                prox_size_control.set_scope(scope)
                global_size_controls.append(prox_size_control)
        else:
            const_size_control = self._model.control_data.create_size_control(
                sizing_type=prime.SizingType.CURVATURE
            )
            const_size_control.set_curvature_sizing_params(
                prime.CurvatureSizingParams(
                    model=self._model,
                    max=constant_size,
                    min=constant_size,
                    growth_rate=growth_rate,
                    normal_angle=normal_angle,
                )
            )
            const_size_control.set_scope(scope)
            global_size_controls.append(const_size_control)
            if is_geodesic or elements_per_gap:
                const_size_control = self._model.control_data.create_size_control(
                    sizing_type=prime.SizingType.CURVATURE
                )
                const_size_control.set_curvature_sizing_params(
                    prime.CurvatureSizingParams(
                        model=self._model,
                        max=constant_size,
                        min=constant_size,
                        growth_rate=growth_rate,
                        normal_angle=normal_angle,
                    )
                )
                const_size_control.set_scope(scope)
                geodesic_global_size_controls.append(const_size_control)

        return min_size, max_size

    def __is_size_field_type_geodesic(self, controls: List[prime.SizeControl]):
        """
        Checks if the size field type is geodesic.
        """
        geodesic = True
        geodesic_types = ["Curvature", "Soft"]
        self._logger.info("Size controls: " + str([control.name for control in controls]))
        for control in controls:
            result = control.get_summary(prime.SizeControlSummaryParams(self._model))
            message = str(result.message)
            ctrl_type = message.split("Type : ")[1]
            ctrl_type = ctrl_type.split("\n")[0]
            self._logger.info("Name: '" + control.name + "' Type: '" + ctrl_type + "'")
            if ctrl_type not in geodesic_types:
                geodesic = False
                self._logger.info("Note: VOLUMETRIC method used.")
                return geodesic
        return geodesic

    def __compute_size_field(
        self, size_controls: List[prime.SizeControl], field: prime.SizeField
    ) -> int:
        """
        Helper method to compute size field.
        """
        self._model.delete_volumetric_size_fields(self._model.get_volumetric_size_fields())
        result = field.compute_volumetric(
            [size_control.id for size_control in size_controls],
            prime.VolumetricSizeFieldComputeParams(model=self._model, enable_multi_threading=False),
        )
        return result.size_field_id

    def __remesh_after_wrap(self, part: prime.Part):
        """
        Surface mesh after wrapping.
        """
        surfer = prime.Surfer(self._model)
        feature_edges = part.get_edge_zonelets_of_label_name_pattern(
            label_name_pattern="___wrapper_feature_path___",
            name_pattern_params=prime.NamePatternParams(model=self._model),
        )
        surf_params = surfer.initialize_surfer_params_for_wrapper()
        surf_params.size_field_type = prime.SizeFieldType.VOLUMETRIC
        surfer.remesh_face_zonelets(
            part_id=part.id,
            face_zonelets=part.get_face_zonelets(),
            edge_zonelets=feature_edges,
            params=surf_params,
        )

    def __create_contact_preventions(
        self,
        contact_prevention_params: List,
        contact_prevention_size: float,
        min_size: float,
        max_size: float,
        growth_rate: float,
        geodesic_global_size_controls: List,
        scope: str,
    ):
        """
        Creates contact prevention controls for all parts.
        Proximity size control created for remeshing.
        """
        all_parts_labels = len(self._model.parts)
        all_parts_labels += sum([len(part.get_labels()) for part in self._model.parts])
        if contact_prevention_size < min_size:
            prime.PrimeRuntimeError("Contact_prevention_size is less than min_size.")
            return
        self._logger.info(
            "Contact prevention enabled for:"
            + str([part.name for part in self._model.parts])
            + str([part.get_labels() for part in self._model.parts if part.get_labels()])
        )

        if all_parts_labels > 1:
            geo_prox_size_control = self._model.control_data.create_size_control(
                sizing_type=prime.SizingType.PROXIMITY
            )
            geo_prox_size_control.set_proximity_sizing_params(
                prime.ProximitySizingParams(
                    model=self._model,
                    elements_per_gap=3.0,
                    max=max_size,
                    min=min_size,
                    growth_rate=growth_rate,
                    ignore_orientation=False,
                )
            )
            geo_prox_size_control.set_scope(scope)
            geodesic_global_size_controls.append(geo_prox_size_control)
            for part in self._model.parts:
                source = prime.ScopeDefinition(model=self._model, part_expression=part.name)
                contact_prevention_params.append(
                    prime.ContactPreventionParams(
                        model=self._model, source_scope=source, size=contact_prevention_size
                    )
                )
                for label in part.get_labels():
                    if part.get_face_zonelets_of_label_name_pattern(
                        label_name_pattern=label,
                        name_pattern_params=prime.NamePatternParams(self._model),
                    ):
                        source = prime.ScopeDefinition(model=self._model, label_expression=label)
                        contact_prevention_params.append(
                            prime.ContactPreventionParams(
                                model=self._model, source_scope=source, size=contact_prevention_size
                            )
                        )
        else:
            self._logger.warning(
                "Contact prevention size specified but \
                insufficient parts and labels identified to define contact."
            )

    def __create_feature_recovery_params(
        self,
        feature_recovery_params: List[prime.FeatureRecoveryParams],
        extract_features: bool,
        create_intersection_loops: bool,
        use_existing_features: bool,
        feature_angle: float,
        enable_feature_octree_refinement: bool,
        scope: prime.ScopeDefinition,
    ):
        features = prime.FeatureExtraction(self._model)
        extracted_features = dict()
        if extract_features:
            face_zonelets_prime_array = self._model.control_data.get_part_zonelets(scope=scope)
            for item in face_zonelets_prime_array:
                res = features.extract_features_on_face_zonelets(
                    part_id=item.part_id,
                    face_zonelets=item.face_zonelets,
                    params=prime.ExtractFeatureParams(
                        model=self._model,
                        feature_angle=feature_angle,
                        separation_angle=20,
                        replace=True,
                        label_name="__extracted__features__",
                    ),
                )
                extracted_features[item.part_id] = res.new_edge_zonelets
            feature_scope = scope
            feature_scope.label_expression = "__extracted__features__"
            extracted_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=feature_scope,
            )
            feature_recovery_params.append(extracted_feature_params)
        if create_intersection_loops:
            for part in self._model.parts:
                others = [item.name for item in self._model.parts if item.name != part.name]
                scope1 = prime.ScopeDefinition(self._model, part_expression=part.name)
                scope2 = prime.ScopeDefinition(self._model, part_expression=','.join(others))
                scope1_zonelets_array = self._model.control_data.get_part_zonelets(scope=scope1)
                scope2_zonelets_array = self._model.control_data.get_part_zonelets(scope=scope2)
                intersect_loops_params = prime.CreateIntersectionEdgeLoopsParams(
                    model=self._model, label_name="__intersect_loops__"
                )
                features.create_intersection_edge_loops(
                    part_face_zonelets=scope1_zonelets_array,
                    intersecting_part_face_zonelets=scope2_zonelets_array,
                    params=intersect_loops_params,
                )
            intersect_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=prime.ScopeDefinition(
                    model=self._model, part_expression="*", label_expression="__intersect_loops__"
                ),
            )
            feature_recovery_params.append(intersect_feature_params)
        if use_existing_features:
            existing_feature_params = prime.FeatureRecoveryParams(
                model=self._model,
                enable_feature_octree_refinement=enable_feature_octree_refinement,
                scope=scope,
            )
            feature_recovery_params.append(existing_feature_params)
        return extracted_features

    def __process_size_fields(
        self,
        global_size_controls: List[prime.SizeControl],
        wrap_size_controls: List[prime.SizeControl],
        geodesic_global_size_controls: List[prime.SizeControl],
        remesh_postwrap: bool,
        use_existing_size_fields: bool,
        recompute_remesh_sizes: bool,
        wrap_params: prime.WrapParams,
        size_fields: List[int],
    ) -> List[int]:

        created_size_fields = []
        field = prime.SizeField(self._model)
        is_geodesic = self.__is_size_field_type_geodesic(global_size_controls + wrap_size_controls)
        if (
            is_geodesic
            and remesh_postwrap
            and not recompute_remesh_sizes
            and not use_existing_size_fields
        ):
            created_size_fields.append(
                self.__compute_size_field(geodesic_global_size_controls + wrap_size_controls, field)
            )

        if not use_existing_size_fields and (global_size_controls + wrap_size_controls):
            if is_geodesic:
                size_control_ids = [i.id for i in (global_size_controls + wrap_size_controls)]
                self._logger.info("Geodesic size controls used for wrap:" + str(size_control_ids))
                wrap_params.sizing_method = prime.SizeFieldType.GEODESIC
                wrap_params.size_control_ids = size_control_ids
            else:
                self._logger.info(
                    "Volumetric size controls used for wrap:"
                    + str([i.name for i in (global_size_controls + wrap_size_controls)])
                )
                field_id = self.__compute_size_field(
                    global_size_controls + wrap_size_controls, field
                )
                wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                wrap_params.size_field_ids = [field_id]
                created_size_fields.append(field_id)
        else:
            if size_fields:
                self._logger.info(
                    "Using size fields provided for wrap:" + str([i for i in size_fields])
                )
                wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                wrap_params.size_field_ids = size_fields
            else:
                size_field_ids = self._model.get_active_volumetric_size_fields()
                if size_field_ids:
                    self._logger.info("Existing size field used for wrap:" + str(size_field_ids))
                    wrap_params.sizing_method = prime.SizeFieldType.VOLUMETRIC
                    wrap_params.size_field_ids = size_field_ids
                else:
                    prime.PrimeRuntimeError("Invalid size fields")
        return created_size_fields

    def __process_after_wrap(
        self,
        remesh_postwrap: bool,
        wrapped_part: prime.Part,
        use_existing_size_fields: bool,
        global_size_controls: List[prime.SizeControl],
        remesh_size_controls: List[prime.SizeControl],
        geodesic_global_size_controls: List[prime.SizeControl],
        wrap_size_controls: List[prime.SizeControl],
        remesh_scope: str,
        recompute_remesh_sizes: bool,
        feature_angle: float,
        wrapper: prime.Wrapper,
    ) -> List[int]:

        created_size_fields = []
        field = prime.SizeField(self._model)
        is_geodesic = self.__is_size_field_type_geodesic(global_size_controls + wrap_size_controls)
        features = prime.FeatureExtraction(self._model)
        if remesh_postwrap and wrapped_part:
            if not use_existing_size_fields and (global_size_controls + remesh_size_controls):
                if recompute_remesh_sizes:

                    if is_geodesic:
                        self._logger.info(
                            "Recomputing sizing after geodesic wrap:"
                            + str(
                                [
                                    i.name
                                    for i in (geodesic_global_size_controls + remesh_size_controls)
                                ]
                            )
                        )
                        for i in range(len(geodesic_global_size_controls)):
                            control = geodesic_global_size_controls[i]
                            control.set_scope(remesh_scope)
                            geodesic_global_size_controls[i] = control
                        created_size_fields.append(
                            self.__compute_size_field(
                                geodesic_global_size_controls + remesh_size_controls, field
                            )
                        )
                    else:
                        self._logger.info(
                            "Recomputing sizing for remesh after volumetric wrap:"
                            + str([i.name for i in (global_size_controls + remesh_size_controls)])
                        )
                        for i in range(len(global_size_controls)):
                            control = global_size_controls[i]
                            control.set_scope(remesh_scope)
                            global_size_controls[i] = control
                        self.__compute_size_field(
                            global_size_controls + remesh_size_controls, field
                        )
                else:
                    if is_geodesic:
                        self._logger.info(
                            "Using geodesic size controls for remesh after geodesic wrap \
                            (consider using recompute_remesh_sizes):"
                            + str(
                                [
                                    i.name
                                    for i in (geodesic_global_size_controls + wrap_size_controls)
                                ]
                            )
                        )

            self.__remesh_after_wrap(wrapped_part)
            wrapper.improve_quality(
                part_id=wrapped_part.id,
                params=prime.WrapperImproveQualityParams(
                    model=self._model, island_count=20, target_skewness=0.9
                ),
            )
            return created_size_fields

    def wrap(
        self,
        min_size: float = None,
        max_size: float = None,
        growth_rate: float = 1.2,
        elements_per_gap: float = None,
        normal_angle: float = 18.0,
        input_parts: str = "*",
        input_labels: str = "*",
        keep_inputs: bool = False,
        region_extract: prime.WrapRegion = prime.WrapRegion.EXTERNAL,
        material_point: List[float] = None,
        extract_features: bool = True,
        create_intersection_loops: bool = False,
        use_existing_features: bool = False,
        enable_feature_octree_refinement: bool = True,
        feature_angle: float = 40.0,
        contact_prevention_size: float = None,
        number_of_threads: int = None,
        remesh_postwrap: bool = True,
        recompute_remesh_sizes: bool = False,
        use_existing_size_fields: bool = False,
        size_fields: List[prime.SizeField] = None,
        wrap_size_controls: List[prime.SizeControl] = None,
        remesh_size_controls: List[prime.SizeControl] = None,
        feature_recovery_params: List[prime.FeatureRecoveryParams] = None,
        contact_prevention_params: List[prime.ContactPreventionParams] = None,
        leak_prevention_params: List[prime.LeakPreventionParams] = None,
    ):
        """Wrap and remesh input.

        Default behaviour is to perform an external wrap of all parts in the model
        using curvature sizing and extracting features.
        The wrap is then remeshed to give a surface mesh for the extracted region.

        Geodesic sizing is used if only soft and curvature controls are set.
        If contact prevention size is set and geodesic sizing is available then contact detection
        is used globally.

        If min_size and max_size is provided, variable size mesh is generated
        between min_size and max_size by applying sizes based on curvature.

        If either of min_size or max_size is provided, constant size mesh
        will be generated with the provided size.

        If neither of min_size or max_size is provided, global min and max settings are used to
        generate a variable size mesh based on curvature.


        Parameters
        ----------
        input_parts : str
            Parts to be wrapped.
            Default = "*"

        input_labels : str
            Labels to be wrapped.
            Default = "*"

        keep_inputs : bool
            Retain inputs.
            Default = False

        region_extract : prime.WrapRegion
            Set region to wrap.
            Default = prime.WrapRegion.EXTERNAL

        material_point : List[float], optional
            Material point needed if region extraction method set to material point.

        min_size: float, optional
            Minimum edge length of the mesh.

        max_size: float, optional
            Maximum edge length of the mesh.

        growth_rate : float
            Default = 1.2

        elements_per_gap : float, optional
            Global proximity size control elements per gap with self proximity.

        normal_angle : float
            Global curvature size control normal angle.
            Default = 18.0

        create_intersection_loops: bool
            Create intersection loops between all parts.
            Default = False

        use_existing_features: bool
            Maintain existing features on parts.
            Default = False

        extract_features: bool
            Extract feature edges using feature angle.
            Default = True

        feature_angle: float
            Angle used to extract features.
            Default = 40.0

        enable_feature_octree_refinement: bool
            Apply refinement to feature edges during wrap.
            Default = True

        contact_prevention_size: float, optional
            Global proximity size controls between all parts.

        number_of_threads: int, optional

        remesh_postwrap : bool
            Remesh wrap.
            Default = True

        recompute_remesh_sizes : bool
            Recompute sizes from global controls using the wrap surface.
            Default = False

        use_existing_size_fields : bool
            Use precomputed size fields.
            Default = False

        size_fields : List[prime.SizeField], optional
            Set size fields to use.

        wrap_size_controls : List[prime.SizeControl], optional
            Set wrap size controls to use.

        remesh_size_controls : List[prime.SizeControl], optional
            Set remesh size controls to use.

        feature_recovery_params : List[prime.FeatureRecoveryParams], optional
            Set feature recovery parameters to use.

        contact_prevention_params : List[prime.ContactPreventionParams], optional
            Set contact prevention parameters to use.

        leak_prevention_params : List[prime.LeakPreventionParams], optional
            Set leak prevention parameters to use.

        Returns
        -------
        Wrapped part
                Returns Part.

        Examples
        --------
        >>> import ansys.meshing.prime as prime
        >>> from ansys.meshing.prime import lucid
        >>> prime_client = prime.launch_prime()
        >>> model = prime_client.model
        >>> mesh = lucid.Mesh(model)
        >>> mesh.read("/my_geometry.stl")
        >>> mesh.wrap(min_size=1, max_size=20, create_intersection_loops=True)
        >>> mesh.write("/mesh_output.pmdat")
        >>> prime_client.exit()

        """
        if size_fields is None:
            size_fields = []
        if wrap_size_controls is None:
            wrap_size_controls = []
        if remesh_size_controls is None:
            remesh_size_controls = []
        if feature_recovery_params is None:
            feature_recovery_params = []
        if contact_prevention_params is None:
            contact_prevention_params = []
        if leak_prevention_params is None:
            leak_prevention_params = []

        global_sf = prime.GlobalSizingParams(model=self._model)
        scope = prime.ScopeDefinition(
            model=self._model,
            part_expression=input_parts,
            label_expression=input_labels,
            entity_type=prime.ScopeEntity.FACEANDEDGEZONELETS,
        )

        # Delete topology and work only only on zonelets for now
        for part in self._model.parts:
            if len(part.get_topo_faces()) > 0:
                part.delete_topo_entities(
                    prime.DeleteTopoEntitiesParams(
                        self._model, delete_geom_zonelets=False, delete_mesh_zonelets=True
                    )
                )

        params = prime.ScopeZoneletParams(model=self._model)
        zonelets = self._model.control_data.get_scope_face_zonelets(scope=scope, params=params)
        part_ids = self._model.control_data.get_scope_parts(scope=scope)

        global_size_controls = []
        geodesic_global_size_controls = []
        computed_size_fields = []

        min_size, max_size = self.__setup_sizing_for_wrapper(
            min_size=min_size,
            max_size=max_size,
            growth_rate=growth_rate,
            normal_angle=normal_angle,
            elements_per_gap=elements_per_gap,
            global_size_controls=global_size_controls,
            geodesic_global_size_controls=geodesic_global_size_controls,
            scope=scope,
            global_sf=global_sf,
            wrap_size_controls=wrap_size_controls,
        )

        wrapper_control = self._model.control_data.create_wrapper_control()
        wrap_params = prime.WrapParams(model=self._model, wrap_region=region_extract)
        if region_extract == prime.WrapRegion.MATERIALPOINT:
            if material_point:
                mp_result = self._model.material_point_data.create_material_point(
                    suggested_name="USER_MP",
                    coords=material_point,
                    params=prime.CreateMaterialPointParams(
                        model=self._model, type=prime.MaterialPointType.LIVE
                    ),
                )
                mp_name = mp_result.assigned_name
                wrapper_control.set_live_material_points([mp_name])
            else:
                prime.PrimeRuntimeError("Error: No material point provided. No wrap done.")
                return None

        wrapper_control.set_geometry_scope(scope)

        # Add contact preventions
        if contact_prevention_size:
            self.__create_contact_preventions(
                contact_prevention_params,
                contact_prevention_size,
                min_size,
                max_size,
                growth_rate,
                geodesic_global_size_controls,
                scope,
            )
        if contact_prevention_params:
            wrapper_control.set_contact_preventions(contact_prevention_params)

        # Add feature recoveries
        extracted_features = self.__create_feature_recovery_params(
            feature_recovery_params=feature_recovery_params,
            extract_features=extract_features,
            create_intersection_loops=create_intersection_loops,
            use_existing_features=use_existing_features,
            feature_angle=feature_angle,
            enable_feature_octree_refinement=enable_feature_octree_refinement,
            scope=scope,
        )
        if feature_recovery_params:
            wrapper_control.set_feature_recoveries(feature_recovery_params)

        # Add leak preventions
        if leak_prevention_params:
            wrapper_control.set_leak_preventions(leak_prevention_params)

        computed_size_fields += self.__process_size_fields(
            global_size_controls=global_size_controls,
            wrap_size_controls=wrap_size_controls,
            geodesic_global_size_controls=geodesic_global_size_controls,
            remesh_postwrap=remesh_postwrap,
            use_existing_size_fields=use_existing_size_fields,
            recompute_remesh_sizes=recompute_remesh_sizes,
            wrap_params=wrap_params,
            size_fields=size_fields,
        )

        # Wrap
        wrapper = prime.Wrapper(model=self._model)
        if number_of_threads:
            wrap_params.number_of_threads = number_of_threads
        res = wrapper.wrap(wrapper_control_id=wrapper_control.id, params=wrap_params)
        wrapped_part = self._model.get_part(res.id)
        remesh_scope = prime.ScopeDefinition(model=self._model, part_expression=wrapped_part.name)

        # Post Wrap Processing
        self.__process_after_wrap(
            remesh_postwrap,
            wrapped_part,
            use_existing_size_fields,
            global_size_controls,
            remesh_size_controls,
            geodesic_global_size_controls,
            wrap_size_controls,
            remesh_scope,
            recompute_remesh_sizes,
            feature_angle,
            wrapper,
        )

        volume_results = wrapped_part.compute_closed_volumes(
            prime.ComputeVolumesParams(
                model=self._model, create_zones_type=prime.CreateVolumeZonesType.PERVOLUME
            )
        )
        self._logger.info(str(volume_results.volumes) + " volumes found.")

        self._model.set_global_sizing_params(
            prime.GlobalSizingParams(
                self._model, min=global_sf.min, max=global_sf.max, growth_rate=global_sf.growth_rate
            )
        )
        self._model.control_data.delete_controls(
            [
                control.id
                for control in [wrapper_control]
                + global_size_controls
                + geodesic_global_size_controls
            ]
        )

        # delete extracted features
        for part_id, zonelets_to_delete in extracted_features.items():
            part = self._model.get_part(part_id)
            part.delete_zonelets(zonelets_to_delete)

        # delete size fields
        if len(computed_size_fields) > 0:
            self._model.delete_volumetric_size_fields(computed_size_fields)

        # retain or delete inputs
        if not keep_inputs:
            self._logger.info("Deleting inputs to wrap.")
            if zonelets:
                for part_id in part_ids:
                    part = self._model.get_part(part_id)
                    [
                        part.delete_zonelets([face])
                        for face in zonelets
                        if face in part.get_face_zonelets()
                    ]
            if part_ids:
                self._model.delete_parts(part_ids)

        self._logger.info("Wrap done.")
        return wrapped_part



import re


def match_pattern(pattern: str, name: str) -> bool:
    pattern = "^" + pattern.replace("*", ".*").replace("?", ".") + "$"
    x = re.search(pattern, name)
    if x:
        return True
    else:
        return False


def check_name_pattern(name_patterns: str, name: str) -> bool:
    patterns = []
    include_pattern = []
    exclude_pattern = []
    a = name_patterns.split(",")
    for aa in a:
        if len(aa) > 0:
            b = aa.split(" ")
            for bb in b:
                if len(bb) > 0:
                    patterns.append(bb)
    for pattern in patterns:
        pattern = pattern.strip()
        if pattern.startswith("!"):
            exclude_pattern.append(pattern[1:])
        else:
            include_pattern.append(pattern)

    for pattern in exclude_pattern:
        if match_pattern(pattern, name):
            return False

    for pattern in include_pattern:
        if match_pattern(pattern, name):
            return True

    return False



from ansys.meshing.prime.autogen.wrappercontrol import WrapperControl as _WrapperControl

# isort: split
from ansys.meshing.prime.autogen.commonstructs import SetNameResults as SetNameResults
from ansys.meshing.prime.internals.comm_manager import CommunicationManager


class WrapperControl(_WrapperControl):
    """Wrapper Control to describe all parameters and controls used for wrapping."""

    def __init__(self, model: CommunicationManager, id: int, object_id: int, name: str):
        """Initialize Part"""
        self._model = model
        _WrapperControl.__init__(self, model, id, object_id, name)

    def __str__(self) -> str:
        return "Not implemented yet"

    def set_suggested_name(self, name: str) -> SetNameResults:
        """Sets the unique name for the wrapper control based on the given suggested name.


        Parameters
        ----------
        name : str
            Suggested name for the wrapper control.

        Returns
        -------
        SetNameResults
            Returns the results with assigned name of the wrapper control.


        Examples
        --------
        >>> wrapper_control.set_suggested_name("wrapper_control1")

        """
        result = _WrapperControl.set_suggested_name(self, name)
        self._name = result.assigned_name
        return result

    @property
    def name(self):
        """Gets the name of wrapper control.

        Returns
        -------
        str
            Returns a name of the wrapper control.

        Examples
        --------
        >>> print(wrapper_control.name)

        """
        return self._name



import enum
import os

import numpy as np
import pyvista as pv
from pyvista import _vtk
from pyvista.plotting.plotting import Plotter

import ansys.meshing.prime as prime
from ansys.meshing.prime.internals import defaults


def compute_face_list_from_structured_nodes(nodes, dim):
    flist = []
    for w in range(dim[2]):
        for u in range(dim[0] - 1):
            for v in range(dim[1] - 1):
                flist.append(4)
                flist.append(u + v * dim[0] + w * dim[0] * dim[1])
                flist.append(u + 1 + v * dim[0] + w * dim[0] * dim[1])
                flist.append(u + 1 + (v + 1) * dim[0] + w * dim[0] * dim[1])
                flist.append(u + (v + 1) * dim[0] + w * dim[0] * dim[1])

    for v in range(dim[1]):
        for u in range(dim[0] - 1):
            for w in range(dim[2] - 1):
                flist.append(4)
                flist.append(u + v * dim[0] + w * dim[0] * dim[1])
                flist.append(u + 1 + v * dim[0] + w * dim[0] * dim[1])
                flist.append(u + 1 + v * dim[0] + (w + 1) * dim[0] * dim[1])
                flist.append(u + v * dim[0] + (w + 1) * dim[0] * dim[1])

    for u in range(dim[0]):
        for v in range(dim[1] - 1):
            for w in range(dim[2] - 1):
                flist.append(4)
                flist.append(u + v * dim[0] + w * dim[0] * dim[1])
                flist.append(u + (v + 1) * dim[0] + w * dim[0] * dim[1])
                flist.append(u + (v + 1) * dim[0] + (w + 1) * dim[0] * dim[1])
                flist.append(u + v * dim[0] + (w + 1) * dim[0] * dim[1])
    return flist


class DisplayMeshType(enum.IntEnum):
    TOPOFACE = 0
    TOPOEDGE = 1
    FACEZONELET = 2
    EDGEZONELET = 3
    SPLINECONTROLPOINTS = 4
    SPLINESURFACE = 5


class ColorByType(enum.IntEnum):
    ZONE = 0
    ZONELET = 1
    PART = 2


''' bright color palette '''
'''color_matrix = np.array([
[0.99609375, 0.76171875, 0.0703125],
[0.96484375, 0.62109375, 0.12109375],
[0.9296875, 0.3515625, 0.140625],
[0.9140625, 0.125, 0.15234375],
[0.765625, 0.89453125, 0.21875],
[0.63671875, 0.79296875, 0.21875],
[0, 0.578125, 0.1953125],
[0, 0.3828125, 0.3984375],
[0.0703125, 0.79296875, 0.765625],
[0.0703125, 0.53515625, 0.65234375],
[0.0234375, 0.3203125, 0.86328125],
[0.10546875, 0.078125, 0.390625],
[0.98828125, 0.65234375, 0.87109375],
[0.84765625, 0.5, 0.9765625],
[0.59765625, 0.5, 0.9765625] ]) * 255 '''


''' winter color palette '''
'''color_matrix = np.array([
[0.2734375, 0.71875, 0.8671875],
[0.2734375, 0.90234375, 0.86328125],
[0.2734375, 0.87109375, 0.62109375],
[0.2734375, 0.80078125, 0.84765625],
[0.2734375, 0.70703125, 0.66796875],
[0.2734375, 0.84375, 0.90234375],
[0.2734375, 0.64453125, 0.71875],
[0.2734375, 0.921875, 0.703125],
[0.2734375, 0.765625, 0.66015625],
[0.2734375, 0.9140625, 0.87890625],
[0.2734375, 0.70703125, 0.5],
[0.2734375, 0.8828125, 0.70703125],
[0.2734375, 0.99609375, 0.8515625],
[0.2734375, 0.91015625, 0.9765625],
[0.2734375, 0.99609375, 0.6171875] ]) * 255'''

''' light fall color palette '''
color_matrix = np.array(
    [
        [155, 186, 126],
        [242, 236, 175],
        [255, 187, 131],
        [194, 187, 97],
        [159, 131, 169],
        [157, 190, 139],
        [233, 218, 158],
        [254, 252, 196],
        [246, 210, 148],
        [215, 208, 198],
        [196, 235, 145],
    ]
)


def compute_distance(point1, point2):
    dist = np.linalg.norm(np.array(point2) - np.array(point1))
    return dist


class Picker:
    def __init__(self, plotter, graphics):
        self.plotter = plotter
        self._graphics = graphics
        self._selected_disp_mesh: list[_DisplayMesh] = []
        self._points = []
        self._ignore = False

    @property
    def selections(self):
        """To access all the selected disp mesh when done."""
        return self._selected_disp_mesh

    def clear_selection(self):
        """ """
        [disp_mesh.deselect() for disp_mesh in self._selected_disp_mesh]
        self._selected_disp_mesh.clear()

    def ignore(self, ignore_pick):
        self._ignore = ignore_pick

    def __call__(self, *args, **kwargs):
        if self._ignore:
            return
        picked_pt = np.array(self.plotter.pick_mouse_position())
        direction = picked_pt - self.plotter.camera_position[0]
        direction = direction / np.linalg.norm(direction)
        start = picked_pt - 1000 * direction
        end = picked_pt + 10000 * direction
        disp_mesh_list = self._graphics.get_face_mesh_data()
        closest_disp_mesh = None
        closest_dist = np.finfo(0.0).max
        for disp_mesh in disp_mesh_list:
            if disp_mesh.poly_data is None:
                continue
            point, ix = disp_mesh.poly_data.ray_trace(start, end, first_point=True)
            if ix.size != 0:
                dist = compute_distance(start, point)
                if dist < closest_dist and disp_mesh not in self._selected_disp_mesh:
                    closest_dist = dist
                    closest_disp_mesh = disp_mesh
                elif disp_mesh in self._selected_disp_mesh:
                    disp_mesh.deselect()
                    self._selected_disp_mesh.remove(disp_mesh)
        if closest_disp_mesh is not None:
            self._selected_disp_mesh.append(closest_disp_mesh)
        [disp_mesh.select() for disp_mesh in self._selected_disp_mesh]
        return


class Graphics(object):
    """ """

    def __init__(self, model: prime.Model):
        """ """
        self._model = model
        self._display_data = {}
        self._display_spline_data = {}
        self._plotter = None
        self._picker = None
        self._color_by_type = ColorByType.ZONE
        self._parts = None
        self._facet_result = prime.FaceAndEdgeConnectivityResults(model=model)
        self._app = None
        self._ruler_visible = False
        self._ruler_actor = None
        self._colorByTypeBt: _vtk.vtkButtonWidget = None
        self._hideBt: _vtk.vtkButtonWidget = None
        self._showEdgeBt: _vtk.vtkButtonWidget = None
        self._printInfoBt: _vtk.vtkButtonWidget = None
        self._showRulerBt: _vtk.vtkButtonWidget = None
        self._sphinx_build = defaults.get_sphinx_build()
        if os.getenv('PRIME_APP_RUN'):
            self._app = __import__('PrimeApp')
        else:
            self.__update_display_data()

    def __update_display_data(self):
        """ """
        part_ids = [part.id for part in self._model.parts]
        mesh_info = prime.MeshInfo(self._model)
        self._display_data.clear()
        with prime.numpy_array_optimization_enabled():
            self._facet_result = mesh_info.get_face_and_edge_connectivity(
                part_ids, prime.FaceAndEdgeConnectivityParams(model=self._model)
            )
        for i, part_id in enumerate(self._facet_result.part_ids):
            data = {}
            part = self._model.get_part(part_id)
            splines = part.get_splines()
            disp_mesh_data: list[_DisplayMesh] = [
                self.__get_face_display_mesh_object(part_id, face_facet_res, j)
                for face_facet_res in self._facet_result.face_connectivity_result_per_part
                for j in range(0, len(face_facet_res.face_zonelet_ids))
            ]
            disp_e_mesh_data: list[_DisplayMesh] = [
                self.__get_edge_display_mesh_object(part_id, edge_facet_res, j)
                for edge_facet_res in self._facet_result.edge_connectivity_result_per_part
                for j in range(0, len(edge_facet_res.edge_zonelet_ids))
            ]
            disp_ctrl_point_data: list[_DisplayMesh] = [
                self.__get_ctrl_point_display_mesh_object(part_ids[i], j) for j in splines
            ]
            disp_spline_surf_data: list[_DisplayMesh] = [
                self.__get_spline_surf_display_mesh_object(part_ids[i], j) for j in splines
            ]
            if len(disp_mesh_data) > 0:
                data["faces"] = disp_mesh_data
            if len(disp_e_mesh_data) > 0:
                data["edges"] = disp_e_mesh_data
            if len(disp_ctrl_point_data) > 0:
                data["ctrlpts"] = disp_ctrl_point_data
            if len(disp_spline_surf_data) > 0:
                data["splinesurf"] = disp_spline_surf_data
            self._display_data[part_id] = data

    def __get_ctrl_point_display_mesh_object(self, part_id, spline_id):
        part = self._model.get_part(part_id)
        spline = part.get_spline(spline_id)
        dim = spline.control_points_count
        nodes = spline.control_points
        face_list = compute_face_list_from_structured_nodes(nodes, dim)
        display_mesh_type = DisplayMeshType.SPLINECONTROLPOINTS
        id = spline.id
        disp_mesh = _DisplayMesh(
            display_mesh_type,
            id,
            part_id,
            self,
            self._model,
            nodes,
            face_list,
            False,
            zone_id=-1,
            zone_name="",
            part_name=part.name,
        )
        return disp_mesh

    def __get_spline_surf_display_mesh_object(self, part_id, spline_id):
        part = self._model.get_part(part_id)
        spline = part.get_spline(spline_id)
        dim = spline.spline_points_count
        nodes = spline.spline_points
        face_list = compute_face_list_from_structured_nodes(nodes, dim)
        display_mesh_type = DisplayMeshType.SPLINESURFACE
        id = spline.id
        disp_mesh = _DisplayMesh(
            display_mesh_type,
            id,
            part_id,
            self,
            self._model,
            nodes,
            face_list,
            False,
            zone_id=-1,
            zone_name="",
            part_name=part.name,
        )
        return disp_mesh

    def __get_face_display_mesh_object(
        self, part_id: int, face_facet_res: prime.FaceConnectivityResults, index: int
    ):
        """ """
        part = self._model.get_part(part_id)
        node_start = 3 * np.sum(face_facet_res.num_nodes_per_face_zonelet[0:index])
        num_node_coords = 3 * face_facet_res.num_nodes_per_face_zonelet[index]
        face_list_start = np.sum(face_facet_res.num_face_list_per_face_zonelet[0:index])
        num_face_list = face_facet_res.num_face_list_per_face_zonelet[index]
        has_mesh = True
        if face_facet_res.topo_face_ids[index] > 0:
            display_mesh_type = DisplayMeshType.TOPOFACE
            id = face_facet_res.topo_face_ids[index]
            has_mesh = face_facet_res.mesh_face_ids[index] > 0
        else:
            display_mesh_type = DisplayMeshType.FACEZONELET
            id = face_facet_res.face_zonelet_ids[index]
        disp_mesh = _DisplayMesh(
            display_mesh_type,
            id,
            part_id,
            self,
            self._model,
            face_facet_res.node_coords[node_start : node_start + num_node_coords].reshape((-1, 3)),
            face_facet_res.face_list[face_list_start : face_list_start + num_face_list],
            has_mesh,
            zone_id=face_facet_res.face_zone_ids[index],
            zone_name=face_facet_res.face_zone_names[index],
            part_name=part.name,
        )
        return disp_mesh

    def __get_edge_display_mesh_object(
        self, part_id: int, edge_facet_res: prime.EdgeConnectivityResults, index: int
    ):
        """ """
        part = self._model.get_part(part_id)
        node_start = 3 * np.sum(edge_facet_res.num_nodes_per_edge_zonelet[0:index])
        num_node_coords = 3 * edge_facet_res.num_nodes_per_edge_zonelet[index]
        edge_list_start = np.sum(edge_facet_res.num_edge_list_per_edge_zonelet[0:index])
        num_edge_list = edge_facet_res.num_edge_list_per_edge_zonelet[index]
        has_mesh = True
        if edge_facet_res.topo_edge_ids[index] > 0:
            display_mesh_type = DisplayMeshType.TOPOEDGE
            id = edge_facet_res.topo_edge_ids[index]
            has_mesh = edge_facet_res.mesh_edge_ids[index] > 0
        else:
            display_mesh_type = DisplayMeshType.EDGEZONELET
            id = edge_facet_res.edge_zonelet_ids[index]
        disp_mesh = _DisplayMesh(
            display_mesh_type,
            id,
            part_id,
            self,
            self._model,
            edge_facet_res.node_coords[node_start : node_start + num_node_coords].reshape((-1, 3)),
            edge_facet_res.edge_list[edge_list_start : edge_list_start + num_edge_list],
            has_mesh,
            part_name=part.name,
            topo_edge_type=edge_facet_res.topo_edge_types[index],
            number_of_edges=edge_facet_res.num_edges_per_edge_zonelet[index],
        )
        return disp_mesh

    def __call__(self, parts=None, update=True, spline=False, scope: prime.ScopeDefinition = None):
        """ """
        self._parts = parts
        # if(spline):
        #     self.__draw_spline(update)
        # el
        if scope != None:
            self.__draw_scope_parts(update, scope)
        elif parts != None:
            self.__draw_parts(parts, update, spline)
        else:
            self.show(update)

    def __color_by_type_callback(self, flag):
        """ """
        vr = self._colorByTypeBt.GetRepresentation()
        state = vr.GetState()
        self._color_by_type = ColorByType(state)
        r = _vtk.vtkPNGReader()
        color_by_type_icon_file = ""
        if self._color_by_type == ColorByType.ZONELET:
            color_by_type_icon_file = os.path.join(
                os.path.dirname(__file__), 'images', 'surface_body.png'
            )
        elif self._color_by_type == ColorByType.ZONE:
            color_by_type_icon_file = os.path.join(os.path.dirname(__file__), 'images', 'bin.png')
        else:
            color_by_type_icon_file = os.path.join(os.path.dirname(__file__), 'images', 'parts.png')
        r.SetFileName(color_by_type_icon_file)
        r.Update()
        image = r.GetOutput()
        vr.SetButtonTexture(state, image)
        [
            disp_mesh.set_color_by_type(self._color_by_type)
            for part_id, data in self._display_data.items()
            if data.get("faces") != None
            for disp_mesh in data["faces"]
        ]

    def __hide_unhide_selection(self, flag):
        """ """
        sel_disp_mesh = self._picker.selections
        if len(sel_disp_mesh) > 0:
            [disp_mesh.hide(self._plotter) for disp_mesh in sel_disp_mesh]
            self._picker.clear_selection()
        else:
            [
                disp_mesh.unhide(self._plotter)
                for part_id, data in self._display_data.items()
                if data.get("faces") != None
                for disp_mesh in data["faces"]
            ]

    def __show_edges_callback(self, flag):
        [
            disp_mesh.show_edges(flag)
            for part_id, data in self._display_data.items()
            if data.get("faces") != None
            for disp_mesh in data["faces"]
        ]

    def __print_callback(self, flag):
        sel_disp_mesh = self._picker.selections
        [print(disp_mesh) for disp_mesh in sel_disp_mesh]

    def __show_ruler_callback(self, flag):
        """This function shows ruler on UI when clicked on ruler button"""
        if self._plotter is not None:
            if self._ruler_visible and self._ruler_actor is not None:
                self._plotter.remove_actor(self._ruler_actor)
                self._ruler_visible = False
            else:
                self._ruler_actor = self._plotter.show_bounds(
                    grid='front',
                    location='outer',
                    all_edges=False,
                    show_xaxis=True,
                    show_yaxis=True,
                    show_zaxis=True,
                )
                self._ruler_visible = True

    def get_face_mesh_data(self):
        """ """
        face_mesh_data = [
            disp_mesh
            for part_id, data in self._display_data.items()
            if data.get("faces") != None
            for disp_mesh in data["faces"]
        ]
        return face_mesh_data

    def show(self, update=False):
        """ """
        if os.getenv('PRIME_APP_RUN') and self._app is not None:
            app_g = self._app.Graphics().Get()
            app_g.DisplayModel()
            app_g.FitToScreen()
            return
        if update == True:
            self.__update_display_data()
        self._plotter = pv.Plotter()
        self._plotter.show_axes()
        [
            disp_mesh.add_to_plotter(self._plotter)
            for part_id, data in self._display_data.items()
            for key, disp_mesh_data in data.items()
            for disp_mesh in disp_mesh_data
        ]
        if self._sphinx_build == False:
            self._colorByTypeBt = self._plotter.add_checkbox_button_widget(
                self.__color_by_type_callback,
                value=int(self._color_by_type) is ColorByType.ZONELET,
                position=(10, 700),
                size=30,
                border_size=3,
            )
            self._hideBt = self._plotter.add_checkbox_button_widget(
                self.__hide_unhide_selection, position=(10, 650), size=30, border_size=3
            )
            self._showEdgeBt = self._plotter.add_checkbox_button_widget(
                self.__show_edges_callback, value=False, position=(10, 600), size=30, border_size=3
            )
            # self._plotter.add_checkbox_button_widget(
            #                  self.__surface_mesh_callback, position=(1870, 700),
            #                  color_on='yellow', color_off='yellow', size = 30, border_size=3)
            self._printInfoBt = self._plotter.add_checkbox_button_widget(
                self.__print_callback, position=(10, 550), size=30, border_size=3
            )
            self._showRulerBt = self._plotter.add_checkbox_button_widget(
                self.__show_ruler_callback, position=(10, 500), size=30, border_size=3
            )
        self._picker = Picker(self._plotter, self)
        self._plotter.track_click_position(self._picker, side='left')
        # self._plotter.window_size = [1920, 1017]
        if self._sphinx_build == False:
            self.__update_bt_icons()
        self._plotter.show()

    def __draw_parts(self, parts=[], update=False, spline=False):
        """ """
        if os.getenv('PRIME_APP_RUN') and self._app is not None:
            app_g = self._app.Graphics().Get()
            app_g.Clear()
            [app_g.DrawPart(part) for part in parts]
            app_g.FitToScreen()
            return
        if update == True:
            self.__update_display_data()
        self._plotter = pv.Plotter()
        self._plotter.show_axes()
        if spline:
            [
                disp_mesh.add_to_plotter(self._plotter)
                for part_id, data in self._display_data.items()
                if (part_id in parts)
                for key, disp_mesh_data in data.items()
                for disp_mesh in disp_mesh_data
            ]
        else:
            [
                disp_mesh.add_to_plotter(self._plotter)
                for part_id, data in self._display_data.items()
                if (part_id in parts)
                for key, disp_mesh_data in data.items()
                for disp_mesh in disp_mesh_data
                if (
                    disp_mesh.type != DisplayMeshType.SPLINECONTROLPOINTS
                    or disp_mesh.type != DisplayMeshType.SPLINESURFACE
                )
            ]
        if self._sphinx_build == False:
            self._colorByTypeBt = self._plotter.add_checkbox_button_widget(
                self.__color_by_type_callback,
                value=self._color_by_type is ColorByType.ZONELET,
                position=(10, 700),
                size=30,
                border_size=3,
            )
            self._hideBt = self._plotter.add_checkbox_button_widget(
                self.__hide_unhide_selection, position=(10, 650), size=30, border_size=3
            )
            self._showEdgeBt = self._plotter.add_checkbox_button_widget(
                self.__show_edges_callback, value=False, position=(10, 600), size=30, border_size=3
            )
            # self._plotter.add_checkbox_button_widget(self.__surface_mesh_callback,
            #                  position=(1870, 700), color_on='yellow',
            #                  color_off='yellow', size = 30, border_size=3)
            self._printInfoBt = self._plotter.add_checkbox_button_widget(
                self.__print_callback, position=(10, 550), size=30, border_size=3
            )
            self._showRulerBt = self._plotter.add_checkbox_button_widget(
                self.__show_ruler_callback, position=(10, 500), size=30, border_size=3
            )
        self._picker = Picker(self._plotter, self)
        self._plotter.track_click_position(self._picker, side='left')
        # self._plotter.window_size = [1920, 1017]
        if self._sphinx_build == False:
            self.__update_bt_icons()
        self._plotter.show()

    def __draw_scope_parts(self, update=False, scope: prime.ScopeDefinition = None):
        """ """
        self._plotter = pv.Plotter()
        self._plotter.show_axes()
        if os.getenv('PRIME_APP_RUN') and self._app is not None:
            app_g = self._app.Graphics().Get()
            app_g.Clear()
        elif update == True:
            self.__update_display_data()

        parts = self._model.control_data.get_scope_parts(scope)
        scope_def = scope
        for part_id in parts:
            part = self._model.get_part(part_id)
            scope_def.part_expression = part.name
            disp_data = None
            disp_ids = []
            if scope.entity_type == prime.ScopeEntity.FACEZONELETS:
                disp_ids = self._model.control_data.get_scope_face_zonelets(
                    scope=scope_def, params=prime.ScopeZoneletParams(model=self._model)
                )
                disp_data = self._display_data[part_id]["faces"]

            if disp_data is not None:
                [
                    disp_mesh.add_to_plotter(self._plotter)
                    for disp_mesh in disp_data
                    if (disp_mesh._id in disp_ids)
                ]

        if self._sphinx_build == False:
            self._colorByTypeBt = self._plotter.add_checkbox_button_widget(
                self.__color_by_type_callback,
                value=self._color_by_type is ColorByType.ZONELET,
                position=(10, 700),
                size=30,
                border_size=3,
            )
            self._hideBt = self._plotter.add_checkbox_button_widget(
                self.__hide_unhide_selection, position=(10, 650), size=30, border_size=3
            )
            self._showEdgeBt = self._plotter.add_checkbox_button_widget(
                self.__show_edges_callback, value=False, position=(10, 600), size=30, border_size=3
            )
            # self._plotter.add_checkbox_button_widget(self.__surface_mesh_callback,
            #                  position=(1870, 700), color_on='yellow',
            #                  color_off='yellow', size = 30, border_size=3)
            self._printInfoBt = self._plotter.add_checkbox_button_widget(
                self.__print_callback, position=(10, 550), size=30, border_size=3
            )
            self._showRulerBt = self._plotter.add_checkbox_button_widget(
                self.__show_ruler_callback, position=(10, 500), size=30, border_size=3
            )
        self._picker = Picker(self._plotter, self)
        self._plotter.track_click_position(self._picker, side='left')
        # self._plotter.window_size = [1920, 1017]
        if self._sphinx_build == False:
            self.__update_bt_icons()
        self._plotter.show()

    def __update_bt_icons(self):
        vr = self._colorByTypeBt.GetRepresentation()
        vr.SetNumberOfStates(3)
        r = _vtk.vtkPNGReader()
        color_by_zone_icon_file = os.path.join(os.path.dirname(__file__), 'images', 'bin.png')
        r.SetFileName(color_by_zone_icon_file)
        r.Update()
        image_1 = r.GetOutput()
        vr.SetButtonTexture(0, image_1)
        hide_vr = self._hideBt.GetRepresentation()
        hide_unhide_icon_file = os.path.join(
            os.path.dirname(__file__), 'images', 'invert_visibility.png'
        )
        hide_r = _vtk.vtkPNGReader()
        hide_r.SetFileName(hide_unhide_icon_file)
        hide_r.Update()
        image_2 = hide_r.GetOutput()
        hide_vr.SetButtonTexture(0, image_2)
        hide_vr.SetButtonTexture(1, image_2)
        show_edge_vr = self._showEdgeBt.GetRepresentation()
        show_edges_icon_file = os.path.join(os.path.dirname(__file__), 'images', 'show_edges.png')
        show_edge_r = _vtk.vtkPNGReader()
        show_edge_r.SetFileName(show_edges_icon_file)
        show_edge_r.Update()
        image_3 = show_edge_r.GetOutput()
        show_edge_vr.SetButtonTexture(0, image_3)
        show_edge_vr.SetButtonTexture(1, image_3)
        print_info_vr = self._printInfoBt.GetRepresentation()
        print_info_icon_file = os.path.join(
            os.path.dirname(__file__), 'images', 'selectioninfo.png'
        )
        print_info_r = _vtk.vtkPNGReader()
        print_info_r.SetFileName(print_info_icon_file)
        print_info_r.Update()
        image_4 = print_info_r.GetOutput()
        print_info_vr.SetButtonTexture(0, image_4)
        print_info_vr.SetButtonTexture(1, image_4)
        show_ruler_vr = self._showRulerBt.GetRepresentation()
        show_ruler_icon_file = os.path.join(os.path.dirname(__file__), 'images', 'show_ruler.png')
        show_ruler_r = _vtk.vtkPNGReader()
        show_ruler_r.SetFileName(show_ruler_icon_file)
        show_ruler_r.Update()
        image_5 = show_ruler_r.GetOutput()
        show_ruler_vr.SetButtonTexture(0, image_5)
        show_ruler_vr.SetButtonTexture(1, image_5)

    def get_color_by_type(self) -> ColorByType:
        """ """
        return self._color_by_type


class _DisplayMesh(object):
    """ """

    def __init__(
        self,
        type: DisplayMeshType,
        id: int,
        part_id: int,
        graphics: Graphics,
        model: prime.Model,
        vertices: np.array,
        facet_list: np.array,
        has_mesh: bool,
        zone_id: int = 0,
        zone_name: str = "",
        part_name: str = "",
        topo_edge_type: int = 0,
        number_of_edges: int = 0,
    ):
        """ """
        self._type = type
        self._id = id
        self._part_id = part_id
        self._geom_id = None
        self._mesh_id = None
        self._graphics = graphics
        self._model = model
        self._vertices = vertices
        self._facet_list = facet_list
        self._n_edges = number_of_edges
        self._topo_edge_type = topo_edge_type
        self._has_mesh = has_mesh
        self._poly_data = None
        self._actor = None
        self._show_edges = self._has_mesh if self._type is DisplayMeshType.TOPOFACE else True
        self._zone_id = zone_id
        self._zone_name = zone_name
        self._part_name = part_name
        # self.__update()

    @property
    def poly_data(self):
        return self._poly_data

    def __str__(self):
        if self._type == DisplayMeshType.TOPOFACE or self._type == DisplayMeshType.FACEZONELET:
            msg = "Selected FaceZonelet "
            if self._type is DisplayMeshType.TOPOFACE:
                msg = "Selected TopoFace "
            msg += str(self._id)
        elif self._type == DisplayMeshType.TOPOEDGE or self._type == DisplayMeshType.EDGEZONELET:
            msg = "Selected EdgeZonelet "
            if self._type is DisplayMeshType.TOPOEDGE:
                msg = "Selected TopoEdge "
            msg += str(self._id)
        msg += ", in Part Id : " + str(self._part_id) + ", Part Name : " + self._part_name + "\n"
        if self._zone_id > 0:
            msg += "Zone Id : " + str(self._zone_id) + ", Zone Name : " + self._zone_name
        return msg

    def add_to_plotter(self, plotter: Plotter):
        """ """
        if self._type == DisplayMeshType.TOPOFACE or self._type == DisplayMeshType.FACEZONELET:
            if self._poly_data == None:
                surf = pv.PolyData(self._vertices, self._facet_list)
                fcolor = np.array(self.get_face_color())
                colors = np.tile(fcolor, (surf.n_faces, 1))
                surf["colors"] = colors
                surf.disp_mesh = self
                self._poly_data = surf
            sh_edge = self._has_mesh if self._type is DisplayMeshType.TOPOFACE else True
            if self._poly_data.n_points > 0:
                self._actor = plotter.add_mesh(
                    self._poly_data, show_edges=sh_edge, scalars="colors", rgb=True, pickable=True
                )
        elif self._type == DisplayMeshType.TOPOEDGE or self._type == DisplayMeshType.EDGEZONELET:
            if self._poly_data == None:
                edge = pv.PolyData()
                edge.points = self._vertices
                cells = np.full((self._n_edges, 3), 2, dtype=np.int_)
                i = 0
                j = 0
                while j < len(self._facet_list):
                    nnodes = self._facet_list[j]
                    j += 1
                    cells[i, 1] = self._facet_list[j]
                    if nnodes == 2:
                        cells[i, 2] = self._facet_list[j + 1]
                    elif nnodes == 3:
                        cells[i, 2] = self._facet_list[j + 2]
                    j += nnodes
                    i += 1
                edge.lines = cells
                ecolor = np.array(self.get_edge_color())
                colors = np.tile(ecolor, (self._n_edges, 1))
                edge["colors"] = colors
                edge.disp_mesh = self
                self._poly_data = edge
            if self._poly_data.n_points > 0:
                self._actor = plotter.add_mesh(
                    self._poly_data, scalars="colors", rgb=True, line_width=4
                )
        elif self._type == DisplayMeshType.SPLINECONTROLPOINTS:
            if self._poly_data == None:
                surf = pv.PolyData(self._vertices, self._facet_list)
                fcolor = np.array([0, 0, 255])
                colors = np.tile(fcolor, (surf.n_faces, 1))
                surf["colors"] = colors
                surf.disp_mesh = self
                self._poly_data = surf
            if self._poly_data.n_points > 0:
                self._actor = plotter.add_mesh(
                    self._poly_data,
                    show_edges=True,
                    scalars="colors",
                    rgb=True,
                    pickable=False,
                    style='wireframe',
                    edge_color=[0, 0, 255],
                )
        elif self._type == DisplayMeshType.SPLINESURFACE:
            if self._poly_data == None:
                surf = pv.PolyData(self._vertices, self._facet_list)
                fcolor = np.array(color_matrix[1])
                colors = np.tile(fcolor, (surf.n_faces, 1))
                surf["colors"] = colors
                surf.disp_mesh = self
                self._poly_data = surf
            if self._poly_data.n_points > 0:
                self._actor = plotter.add_mesh(
                    self._poly_data, show_edges=False, scalars="colors", rgb=True, pickable=False
                )

    def get_face_color(self):
        """ """
        type = self._graphics.get_color_by_type()
        num_colors = int(color_matrix.size / 3)
        if type == ColorByType.ZONELET:
            return color_matrix[self._id % num_colors].tolist()
        elif type == ColorByType.PART:
            return color_matrix[self._part_id % num_colors].tolist()
        else:
            return color_matrix[self._zone_id % num_colors].tolist()

    def get_edge_color(self):
        """ """
        num_colors = int(color_matrix.size / 3)
        if self._type == DisplayMeshType.EDGEZONELET:
            return color_matrix[self._id % num_colors].tolist()
        elif self._type == DisplayMeshType.TOPOEDGE:
            if self._topo_edge_type == 1:
                return [255, 0, 0]
            elif self._topo_edge_type == 2:
                return [0, 0, 0]
            elif self._topo_edge_type == 3:
                return [0, 255, 255]
            elif self._topo_edge_type == 4:
                return [255, 0, 255]
            elif self._topo_edge_type == 5:
                return [255, 255, 0]
            elif self._topo_edge_type == 6:
                return [128, 0, 128]
            else:
                return color_matrix[self._id % num_colors].tolist()

    def deselect(self):
        if self._type == DisplayMeshType.TOPOFACE or self._type == DisplayMeshType.FACEZONELET:
            if self._poly_data != None:
                fcolor = np.array(self.get_face_color())
                colors = np.tile(fcolor, (self._poly_data.n_faces, 1))
                self._poly_data["colors"] = colors

    def select(self):
        if self._type == DisplayMeshType.TOPOFACE or self._type == DisplayMeshType.FACEZONELET:
            if self._poly_data != None:
                # fcolor = np.array([255, 255, 0])
                fcolor = np.array([233, 99, 28])
                colors = np.tile(fcolor, (self._poly_data.n_faces, 1))
                self._poly_data["colors"] = colors

    def hide(self, plotter: Plotter):
        plotter.remove_actor(self._actor)
        self._actor = None

    def clear(self, plotter: Plotter):
        plotter.remove_actor(self._actor)
        del self._actor
        self._actor = None

    def unhide(self, plotter: Plotter):
        if self._actor is None and self._poly_data.n_points > 0:
            self._actor = plotter.add_mesh(
                self._poly_data,
                show_edges=self._show_edges,
                scalars="colors",
                rgb=True,
                pickable=True,
            )

    def show_edges(self, show: bool):
        if self._actor is not None and self._has_mesh:
            prop = self._actor.GetProperty()
            self._show_edges = not prop.GetEdgeVisibility()
            prop.SetEdgeVisibility(not prop.GetEdgeVisibility())

    def set_color_by_type(self, type: ColorByType):
        """ """
        if self._type == DisplayMeshType.TOPOFACE or self._type == DisplayMeshType.FACEZONELET:
            if self._poly_data != None:
                fcolor = np.array(self.get_face_color())
                colors = np.tile(fcolor, (self._poly_data.n_faces, 1))
                self._poly_data["colors"] = colors

    @property
    def type(self):
        return self._type



__version__ = '1.0.0'

__all__ = ['loads', 'dumps', 'JSONEncoder', 'JSONDecoder', 'JSONDecodeError']

from .decoder import JSONDecodeError, JSONDecoder
from .encoder import JSONEncoder

_default_encoder = JSONEncoder(
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    separators=None,
    default=None,
)

_default_decoder = JSONDecoder(object_hook=None, object_pairs_hook=None)


def dumps(
    obj,
    *,
    skipkeys=False,
    ensure_ascii=True,
    check_circular=True,
    allow_nan=True,
    cls=None,
    separators=None,
    default=None,
    sort_keys=False,
    **kw,
):
    """Serialize ``obj`` to a JSON formatted ``str``.

    If ``skipkeys`` is true then ``dict`` keys that are not basic types
    (``str``, ``int``, ``float``, ``bool``, ``None``) will be skipped
    instead of raising a ``TypeError``.

    If ``ensure_ascii`` is false, then the return value can contain non-ASCII
    characters if they appear in strings contained in ``obj``. Otherwise, all
    such characters are escaped in JSON strings.

    If ``check_circular`` is false, then the circular reference check
    for container types will be skipped and a circular reference will
    result in an ``RecursionError`` (or worse).

    If ``allow_nan`` is false, then it will be a ``ValueError`` to
    serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``) in
    strict compliance of the JSON specification, instead of using the
    JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

    If specified, ``separators`` should be an ``(item_separator, key_separator)``
    tuple.  The default is ``(', ', ': ')`` if *indent* is ``None`` and
    ``(',', ': ')`` otherwise.  To get the most compact JSON representation,
    you should specify ``(',', ':')`` to eliminate whitespace.

    ``default(obj)`` is a function that should return a serializable version
    of obj or raise TypeError. The default simply raises TypeError.

    If *sort_keys* is true (default: ``False``), then the output of
    dictionaries will be sorted by key.

    To use a custom ``JSONEncoder`` subclass (e.g. one that overrides the
    ``.default()`` method to serialize additional types), specify it with
    the ``cls`` kwarg; otherwise ``JSONEncoder`` is used.

    """
    # cached encoder
    if (
        not skipkeys
        and ensure_ascii
        and check_circular
        and allow_nan
        and cls is None
        and separators is None
        and default is None
        and not sort_keys
        and not kw
    ):
        return _default_encoder.encode(obj)
    if cls is None:
        cls = JSONEncoder
    return cls(
        skipkeys=skipkeys,
        ensure_ascii=ensure_ascii,
        check_circular=check_circular,
        allow_nan=allow_nan,
        separators=separators,
        default=default,
        sort_keys=sort_keys,
        **kw,
    ).encode(obj)


def loads(
    s: bytes or bytearray,
    *,
    cls=None,
    object_hook=None,
    parse_int=None,
    parse_float=None,
    parse_constant=None,
    object_pairs_hook=None,
    **kw,
):
    if isinstance(s, str):
        import json

        return json.loads(
            s,
            cls=cls,
            object_hook=object_hook,
            parse_int=parse_int,
            parse_float=parse_float,
            parse_constant=parse_constant,
            object_pairs_hook=object_pairs_hook,
            **kw,
        )
    if not isinstance(s, (bytes, bytearray)):
        raise TypeError(
            f'the JSON object must be str, bytes or bytearray, ' f'not {s.__class__.__name__}'
        )

    if (
        cls is None
        and object_hook is None
        and parse_int is None
        and parse_float is None
        and parse_constant is None
        and object_pairs_hook is None
        and not kw
    ):
        return _default_decoder.decode(s)
    if cls is None:
        cls = JSONDecoder
    if object_hook is not None:
        kw['object_hook'] = object_hook
    if object_pairs_hook is not None:
        kw['object_pairs_hook'] = object_pairs_hook
    if parse_float is not None:
        kw['parse_float'] = parse_float
    if parse_int is not None:
        kw['parse_int'] = parse_int
    if parse_constant is not None:
        kw['parse_constant'] = parse_constant
    return cls(**kw).decode(s)



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class AutoMesh(CoreObject):
    """Generates volume mesh.

    Performs volume meshing using various volume meshing algorithms.
    For example, with prisms.
    """

    def __init__(self, model: CommunicationManager):
        """ Initialize AutoMesh """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::AutoMesh/Construct"
        args = {"ModelID" : model._object_id , "MaxID" : -1 }
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for AutoMesh. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for AutoMesh. """
        command_name = "PrimeMesh::AutoMesh/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def mesh(self, part_id : int, automesh_params : AutoMeshParams) -> AutoMeshResults:
        """ Performs volume meshing on the part with the given meshing parameters.


        Parameters
        ----------
        part_id : int
            Id of the part.
        automesh_params : AutoMeshParams
            Parameters for auto mesh.

        Returns
        -------
        AutoMeshResults
            Returns the AutomeshResults.


        Examples
        --------
        >>> auto_mesh = AutoMesh(model = model)
        >>> automesh_params = AutoMeshParams(model = model)
        >>> results = auto_mesh.mesh(part_id, automesh_params)

        """
        args = {"part_id" : part_id,
        "automesh_params" : automesh_params._jsonify()}
        command_name = "PrimeMesh::AutoMesh/Mesh"
        self._model._print_logs_before_command("mesh", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("mesh", AutoMeshResults(model = self._model, json_data = result))
        return AutoMeshResults(model = self._model, json_data = result)



import logging
import os

import ansys.meshing.prime.examples as examples
import ansys.meshing.prime.internals.config as config
import ansys.meshing.prime.internals.defaults as defaults
import ansys.meshing.prime.internals.utils as utils
from ansys.meshing.prime.core.model import Model
from ansys.meshing.prime.internals.utils import terminate_process

__all__ = ['Client']


class Client(object):
    '''Client class for PyPrimeMesh.'''

    def __init__(
        self,
        *,
        server_process=None,
        ip: str = defaults.ip(),
        port: int = defaults.port(),
        timeout: float = defaults.connection_timeout(),
        credentials=None,
        **kwargs,
    ):
        self._default_model: Model = None
        local = kwargs.get('local', False)
        if local and server_process is not None:
            raise ValueError('Local client cannot be instantiated with a server process')
        self._local = local
        self._process = server_process
        self._comm = None
        if not local:
            try:
                from ansys.meshing.prime.internals.grpc_communicator import (
                    GRPCCommunicator,
                )

                channel = kwargs.get('channel', None)
                if channel is not None:
                    self._comm = GRPCCommunicator(channel=channel, timeout=timeout)
                else:
                    self._comm = GRPCCommunicator(
                        ip=ip, port=port, timeout=timeout, credentials=credentials
                    )
                    setattr(self, 'port', port)
            except ImportError as err:
                logging.getLogger('PyPrimeMesh').error(
                    f'Failed to load grpc_communicator with message: {err.msg}'
                )
                raise
            except ConnectionError:
                self.exit()

                logging.getLogger('PyPrimeMesh').error('Failed to connect to PRIME GRPC server')
                raise
        else:
            try:
                from ansys.meshing.prime.internals.prime_communicator import (
                    PrimeCommunicator,
                )

                self._comm = PrimeCommunicator()
            except ImportError as err:
                logging.getLogger('PyPrimeMesh').error(
                    f'Failed to load prime_communicator with message: {err.msg}'
                )

    @property
    def model(self):
        '''Gets model associated with the client.'''
        if self._default_model is None:
            # This assumes that the Model is always object id 1....
            self._default_model = Model(self._comm, 1, 1, "Default")
        return self._default_model

    def run_on_server(self, recipe: str):
        '''Run a recipe on server.

        Parameters
        ----------
        recipe: str
            Recipe to run on the server. This needs to be a valid
            python script.
        '''
        if self._comm is not None:
            result = self._comm.run_on_server(recipe)
            return result['Results']

    def exit(self):
        '''Close the connection with the server.

        If the client had launched the server, then this will also
        kill the server process.

        Examples
        --------
        >>> import ansys.meshing.prime as prime
        >>> prime_client = prime.launch_prime() # This will launch a server process
        >>> model = prime_client.model
        >>> fileio = prime.FileIO(model)
        >>> result = fileio.read_pmdat('example.pmdat', prime.FileReadParams(model=model))
        >>> print(result)
        >>> prime_client.exit() # Sever connection with server and kill server
        '''
        if self._comm is not None:
            self._comm.close()
            self._comm = None
        if self._process is not None:
            assert self._local == False
            terminate_process(self._process)
            self._process = None

        if config.using_container():
            container_name = getattr(self, 'container_name')
            utils.stop_prime_github_container(container_name)
        elif config.has_pim():
            self.remote_instance.delete()
            self.pim_client.close()

        clear_examples = bool(int(os.environ.get('PYPRIMEMESH_CLEAR_EXAMPLES', '1')))
        if clear_examples:
            examples.clear_download_cache()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.exit()



""" Auto-generated file. DO NOT MODIFY """
from __future__ import annotations
from ansys.meshing.prime.internals.comm_manager import CommunicationManager
from ansys.meshing.prime.params.primestructs import *
from ansys.meshing.prime.autogen.coreobject import *
from typing import List, Any, Union

class SurfaceSearch(CoreObject):
    """SurfaceSearch allows you to check surface mesh quality.

    SurfaceSearch performs surface mesh quality check based on different face quality measures.
    """

    def __init__(self, model: CommunicationManager):
        """ Initialize SurfaceSearch """
        self._model = model
        self._comm = model._communicator
        command_name = "PrimeMesh::SurfaceSearch/Construct"
        args = {"ModelID" : model._object_id , "MaxID" : -1 }
        result = self._comm.serve(model, command_name, args=args)
        self._object_id = result["ObjectIndex"]
        self._freeze()

    def __enter__(self):
        """ Enter context for SurfaceSearch. """
        return self

    def __exit__(self, type, value, traceback) :
        """ Exit context for SurfaceSearch. """
        command_name = "PrimeMesh::SurfaceSearch/Destruct"
        self._comm.serve(self._model, command_name, self._object_id, args={})

    def search_zonelets_by_quality(self, part_id : int, face_zonelets : Iterable[int], register_id : int, params : SearchByQualityParams) -> SearchByQualityResults:
        """ Search face zonelets with the provided quality parameters and applies register id on face elements found.


        Parameters
        ----------
        part_id : int
            Id of the part.
        face_zonelets : Iterable[int]
            Ids of the face zonelets whose element search is performed.
        register_id : int
            Id used to register face elements found by the search.
        params : SearchByQualityParams
            Quality parameters used to search face elements.

        Returns
        -------
        SearchByQualityResults
            Returns the SearchByQualityResults.

        Examples
        --------
        >>> surf_search = prime.SurfaceSearch(model=model)
        >>> results = surf_search.search_zonelets_by_quality(part_id, face_zonelets, register_id, prime.SearchByQualityParams(model=model))

        """
        args = {"part_id" : part_id,
        "face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/SearchZoneletsByQuality"
        self._model._print_logs_before_command("search_zonelets_by_quality", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("search_zonelets_by_quality", SearchByQualityResults(model = self._model, json_data = result))
        return SearchByQualityResults(model = self._model, json_data = result)

    def search_zonelets_by_self_intersections(self, part_id : int, face_zonelets : Iterable[int], register_id : int, params : SearchBySelfIntersectionParams) -> SearchByIntersectionResults:
        """ Search face zonelets to identify face elements intersecting with each other.


        Parameters
        ----------
        part_id : int
            Id of part.
        face_zonelets : Iterable[int]
            Ids of face zonelets.
        register_id : int
            Id of register.
        params : SearchBySelfIntersectionParams
            Parameters used to identify face elements by self intersection.

        Returns
        -------
        SearchByIntersectionResults
            Returns the SearchByIntersectionResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.search_zonelets_by_self_intersections(part_id, face_zonelets, register_id, params)

        """
        args = {"part_id" : part_id,
        "face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/SearchZoneletsBySelfIntersections"
        self._model._print_logs_before_command("search_zonelets_by_self_intersections", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("search_zonelets_by_self_intersections", SearchByIntersectionResults(model = self._model, json_data = result))
        return SearchByIntersectionResults(model = self._model, json_data = result)

    def search_zonelets_by_spikes(self, part_id : int, face_zonelets : Iterable[int], register_id : int, params : SearchBySpikeParams) -> SearchBySpikeResults:
        """ Search face zonelets to identify spikes.

        Search face zonelets to identify spikes with provided spike parameters

        Parameters
        ----------
        part_id : int
            Id of part.
        face_zonelets : Iterable[int]
            Ids of face zonelets.
        register_id : int
            Id of register.
        params : SearchBySpikeParams
            Parameters used to identify spikes.

        Returns
        -------
        SearchBySpikeResults
            Returns the SearchBySpikeResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.search_zonelets_by_spikes(part_id, face_zonelets, register_id, params)

        """
        args = {"part_id" : part_id,
        "face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/SearchZoneletsBySpikes"
        self._model._print_logs_before_command("search_zonelets_by_spikes", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("search_zonelets_by_spikes", SearchBySpikeResults(model = self._model, json_data = result))
        return SearchBySpikeResults(model = self._model, json_data = result)

    def search_zonelets_by_folds(self, part_id : int, face_zonelets : Iterable[int], register_id : int, params : SearchByFoldsParams) -> SearchByFoldsResults:
        """ Search face zonelets to identify folds.

        Search face zonelets to identify folds with provided folds parameters.

        Parameters
        ----------
        part_id : int
            Id of part.
        face_zonelets : Iterable[int]
            Ids of face zonelets.
        register_id : int
            Id of register.
        params : SearchByFoldsParams
            Parameters used to identify folds.

        Returns
        -------
        SearchByFoldsResults
            Returns the SearchByFoldsResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.search_zonelets_by_folds(part_id, face_zonelets, register_id, params)

        """
        args = {"part_id" : part_id,
        "face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/SearchZoneletsByFolds"
        self._model._print_logs_before_command("search_zonelets_by_folds", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("search_zonelets_by_folds", SearchByFoldsResults(model = self._model, json_data = result))
        return SearchByFoldsResults(model = self._model, json_data = result)

    def search_zonelets_by_thin_strips(self, part_id : int, face_zonelets : Iterable[int], register_id : int, params : SearchByThinStripParams) -> SearchByThinStripResults:
        """ Search face zonelets to identify face element of thin strips(single layer of triangles between features).


        Parameters
        ----------
        part_id : int
            Id of part.
        face_zonelets : Iterable[int]
            Ids of face zonelets.
        register_id : int
            Id of register.
        params : SearchByThinStripParams
            Parameters used to identify face elements of thin strips.

        Returns
        -------
        SearchByThinStripResults
            Returns the SearchByThinStripResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.search_zonelets_by_thin_strips(part_id, face_zonelets, register_id, params)

        """
        args = {"part_id" : part_id,
        "face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/SearchZoneletsByThinStrips"
        self._model._print_logs_before_command("search_zonelets_by_thin_strips", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("search_zonelets_by_thin_strips", SearchByThinStripResults(model = self._model, json_data = result))
        return SearchByThinStripResults(model = self._model, json_data = result)

    def get_surface_quality_summary(self, params : SurfaceQualitySummaryParams) -> SurfaceQualitySummaryResults:
        """ Gets the surface quality summary.

        Diagnose surface quality for the given scope and face quality measures provided by the surface quality summary parameters.
        Uses default quality limit if not specified with params.

        Parameters
        ----------
        params : SurfaceQualitySummaryParams
            Surface quality summary parameters.

        Returns
        -------
        SurfaceQualitySummaryResults
            Returns the SurfaceQualitySummaryResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.get_surface_quality_summary(SurfaceQualitySummaryParams(model=model))

        """
        args = {"params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/GetSurfaceQualitySummary"
        self._model._print_logs_before_command("get_surface_quality_summary", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("get_surface_quality_summary", SurfaceQualitySummaryResults(model = self._model, json_data = result))
        return SurfaceQualitySummaryResults(model = self._model, json_data = result)

    def get_surface_diagnostic_summary(self, params : SurfaceDiagnosticSummaryParams) -> SurfaceDiagnosticSummaryResults:
        """ Gets the surface diagnostic summary.

        Diagnose surface connectivity for the given scope and controls provided by the surface diagnostic summary parameters.

        Parameters
        ----------
        params : SurfaceDiagnosticSummaryParams
            Surface diagnostic summary parameters.

        Returns
        -------
        SurfaceDiagnosticSummaryResults
            Returns the SurfaceDiagnosticSummaryResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> results = surf_search.get_surface_diagnostics_summary(SurfaceDiagnosticSummaryParams(model=model))

        """
        args = {"params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/GetSurfaceDiagnosticSummary"
        self._model._print_logs_before_command("get_surface_diagnostic_summary", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("get_surface_diagnostic_summary", SurfaceDiagnosticSummaryResults(model = self._model, json_data = result))
        return SurfaceDiagnosticSummaryResults(model = self._model, json_data = result)

    def get_search_info_by_register_id(self, face_zonelets : Iterable[int], register_id : int, params : SearchInfoByRegisterIdParams) -> SearchInfoByRegisterIdResults:
        """ Gets search information regarding registered face elements of provided zonelets using a register id.


        Parameters
        ----------
        face_zonelets : Iterable[int]
            Ids of the face zonelets to search in.
        register_id : int
            An integer register id.
        params : SearchInfoByRegisterIdParams
            Parameters for retrieveing information on registered faces.

        Returns
        -------
        SearchInfoByRegisterIdResults
            Returns the SearchInfoByRegisterIdResults.

        Examples
        --------
        >>> surf_search = SurfaceSearch(model=model)
        >>> params = prime.SearchInfoByRegisterIdParams(model=model)
        >>> results = surf_search.get_search_info_by_register_id(face_zonelets, register_id, params)

        """
        args = {"face_zonelets" : face_zonelets,
        "register_id" : register_id,
        "params" : params._jsonify()}
        command_name = "PrimeMesh::SurfaceSearch/GetSearchInfoByRegisterId"
        self._model._print_logs_before_command("get_search_info_by_register_id", args)
        result = self._comm.serve(self._model, command_name, self._object_id, args=args)
        self._model._print_logs_after_command("get_search_info_by_register_id", SearchInfoByRegisterIdResults(model = self._model, json_data = result))
        return SearchInfoByRegisterIdResults(model = self._model, json_data = result)



from abc import abstractmethod


class Communicator(object):
    @abstractmethod
    def serve(self, model, command, *args, **kwargs) -> dict:
        pass

    @abstractmethod
    def initialize_params(self, model, param_name: str, *args) -> dict:
        pass

    @abstractmethod
    def run_on_server(self, model, recipe: str) -> dict:
        pass

    @abstractmethod
    def import_cad(self, model, file_name: str, *args) -> dict:
        pass

    @abstractmethod
    def close(self):
        pass
